from .NN_LinearLogit_Sklearn import NN_LinearLogit_Sklearn
from .NN_LogitLogit_Sklearn import NN_LogitLogit_Sklearn
from .PCLR import PCLR
from .PCR import PCR
from .PLS_SIMPLS import PLS_SIMPLS
from .RF import RF
from .SVM import SVM
from .NN_L1 import NN_L1
from .NN_L2 import NN_L2
from .NN_LinearLinear import NN_LinearLinear
from .NN_LinearLogit import NN_LinearLogit
from .NN_LinearTan import NN_LinearTan
from .NN_LinearRelu import NN_LinearRelu
from .NN_LogitLinear import NN_LogitLinear
from .NN_LogitLogit import NN_LogitLogit
from .NN_LogitTan import NN_LogitTan
from .NN_LogitRelu import NN_LogitRelu
from .NN_TanLinear import NN_TanLinear
from .NN_TanLogit import NN_TanLogit
from .NN_TanTan import NN_TanTan
from .NN_TanRelu import NN_TanRelu
from .NN_ReluLinear import NN_ReluLinear
from .NN_ReluLogit import NN_ReluLogit
from .NN_ReluTan import NN_ReluTan
from .NN_ReluRelu import NN_ReluRelu


__all__ = ["NN_LinearLogit_Sklearn", "NN_LogitLogit_Sklearn", "PCLR", "PCR", "PLS_SIMPLS", "RF", "SVM", "NN_L1", "NN_L2",
           "NN_LinearLinear", "NN_LinearLogit", "NN_LinearTan", "NN_LinearRelu", " NN_LogitLinear", "NN_LogitLogit",
           "NN_LogitTan", "NN_LogitRelu",
           "NN_TanLinear", "NN_TanLogit", "NN_TanTan", "NN_TanRelu",
           "NN_ReluLinear", "NN_ReluLogit", "NN_ReluTan", "NN_ReluRelu"]



import numpy as np
from sklearn import preprocessing


def color_scale(x, method="tanh", beta=None):

    # Initially scale between 0 and 1
    scaler = preprocessing.MinMaxScaler(feature_range=(0.1, 1))
    x_init = scaler.fit_transform(x[:, np.newaxis]).flatten()

    # Methods of transformation
    if method == "linear":
        x_tr = x_init
    elif method == "sq":
        x_tr = x_init ** 2
    elif method == "sqrt":
        x_tr = np.sqrt(x_init)
    elif method == "tan":
        x_tr = 1 + np.tan(beta * (1 + x_init))
    elif method == "tanh+1":
        x_tr = 1 + np.tanh(beta * (-1 + x_init))
    elif method == "tanh":
        x_tr_init = np.tanh(beta * (-1 + x_init))
        x_tr = scaler.fit_transform(x_tr_init[:, np.newaxis]).flatten()
    else:
        print("An incorrect method for color_scale was selected, so it set to 'tanh'. Supported methods are 'linear', 'sq', 'sqrt', 'tanh', and 'tanh+1'.")
        x_tr = x_init

    return x_tr



import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.linear_model import LogisticRegression
from .BaseModel import BaseModel


class PCLR(BaseModel):
    """ Principal component logistic regression.

    Parameters
    ----------
    n_components : int, (default 2)
        Number of components to keep.

    Methods
    -------
    train : Fit model to data.

    test : Apply model to test data.

    evaluate : Evaluate model.

    calc_bootci : Calculate bootstrap intervals for plot_featureimportance.

    plot_featureimportance : Plot coefficient and Variable Importance in Projection (VIP).

    plot_permutation_test : Perform a permutation test and plot.
    """

    parametric = True
    bootlist = ["model.coef_"]  # list of metrics to bootstrap

    def __init__(self, n_components=2):
        self.model = PCA(n_components=n_components)
        self.regrmodel = LogisticRegression(solver="liblinear")
        self.k = n_components

    def set_params(self, params):
        self.__init__(**params)

    def train(self, X, Y):
        """ Fit the PCR model, save additional stats (as attributes) and return Y predicted values.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.

        Y : array-like, shape = [n_samples, 1]
            Response variables, where n_samples is the number of samples.

        Returns
        -------
        y_pred_train : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # Ensure array and error check
        X, Y = self.input_check(X, Y)

        # Fit the model
        self.model.fit(X)
        self.model.x_scores_ = self.model.transform(X)
        self.regrmodel.fit(self.model.x_scores_, Y)

        # Save x_loadings, coef, pctvar, x_weights, y_loadings and vip
        self.model.x_loadings_ = self.model.components_.T
        self.model.coef_ = np.dot(self.model.x_loadings_, self.regrmodel.coef_.flatten())
        self.model.pctvar_ = self.model.explained_variance_
        self.model.x_weights_ = self.model.components_.T
        self.model.y_loadings_ = self.regrmodel.coef_.reshape(1, len(self.regrmodel.coef_.flatten()))

        # Calculate and return Y prediction value
        y_pred_train = self.regrmodel.predict(self.model.x_scores_).flatten()

        # Storing X, Y, and Y_pred
        self.X = X
        self.Y = Y
        self.Y_pred = y_pred_train
        return y_pred_train

    def test(self, X):
        """Calculate and return Y predicted value.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Test variables, where n_samples is the number of samples and n_features is the number of predictors.

        Returns
        -------
        y_pred_test : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # Convert to X to numpy array if a DataFrame
        if isinstance(X, pd.DataFrame or pd.Series):
            X = np.array(X)

        # Calculate and return Y predicted value
        newX = self.model.transform(X)
        y_pred_test = self.regrmodel.predict(newX).flatten()
        return y_pred_test



import numpy
import numpy as np
import math
from sklearn.cluster import KMeans
from .BaseModel import BaseModel


class RBF_NN(BaseModel):
    """Radial basis function neural network"""

    parametric = True
    bootlist = None

    def __init__(self, n_clusters=8, max_iter=100):
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.k = n_clusters

    def set_params(self, params):
        self.__init__(**params)

    def train(self, X, Y):
        """ Fit the rbf-nn model, save additional stats (as attributes) and return Y predicted values.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.

        Y : array-like, shape = [n_samples, 1]
            Response variables, where n_samples is the number of samples.

        Returns
        -------
        y_pred_train : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # Ensure array and error check
        X, Y = self.input_check(X, Y)

        km = KMeans(n_clusters=self.n_clusters, max_iter=self.max_iter)
        km.fit(X)
        cent = km.cluster_centers_

        self.model = KMeans

        # Determine the value of sigma
        max = 0
        for i in range(self.n_clusters):
            for j in range(self.n_clusters):
                d = numpy.linalg.norm(cent[i] - cent[j])
                if d > max:
                    max = d
        d = max
        sigma = d / math.sqrt(2 * self.n_clusters)

        # Set up G matrix
        shape = X.shape
        row = shape[0]
        column = self.n_clusters
        G = numpy.empty((row, column), dtype=float)
        for i in range(row):
            for j in range(column):
                dist = numpy.linalg.norm(X[i] - cent[j])
                G[i][j] = math.exp(-math.pow(dist, 2) / math.pow(2 * sigma, 2))

        # Find W
        GTG = numpy.dot(G.T, G)
        GTG_inv = numpy.linalg.inv(GTG)
        fac = numpy.dot(GTG_inv, G.T)
        W = numpy.dot(fac, Y)
        self.cent = cent
        self.W = W
        self.G = G
        self.sigma = sigma
        y_pred_train = np.dot(G, W)
        self.xcols_num = len(X.T)
        cent2 = []
        for i in range(len(self.cent.T)):
            something = []
            for j in range(len(self.cent)):
                something.append(self.cent[j][i])
            cent2.append(something)

        self.cent2 = np.array(cent2)

        self.vi = np.dot(self.cent2, self.W)

        self.model.x_scores_ = self.G
        self.model.y_loadings_ = self.W.reshape(1, len(self.W))
        self.model.pctvar_ = np.ones((1, len(self.model.y_loadings_[0])))
        self.X = X
        self.Y = Y
        self.Y_pred = y_pred_train
        return y_pred_train

    def test(self, X):
        """Calculate and return Y predicted value.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Test variables, where n_samples is the number of samples and n_features is the number of predictors.

        Returns
        -------
        y_pred_test : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # Set up G matrix
        shape = X.shape
        row = shape[0]
        column = self.n_clusters
        G = numpy.empty((row, column), dtype=float)
        for i in range(row):
            for j in range(column):
                dist = numpy.linalg.norm(X[i] - self.cent[j])
                G[i][j] = math.exp(-math.pow(dist, 2) / math.pow(2 * self.sigma, 2))
        y_pred_test = numpy.dot(G, self.W)
        return y_pred_test



import numpy as np


def dict_std(dict_list):
    std_dict = {}
    for key in dict_list[0].keys():
        value = []
        for i in dict_list:
            value.append(i[key])
        std_dict[key] = np.std(value, ddof=1)
    return std_dict



import numpy as np
from keras.callbacks import Callback
from keras.optimizers import SGD
from keras.models import Sequential
from keras.layers import Dense
from keras import regularizers
from .BaseModel import BaseModel
from ..utils import YpredCallback


class NN_L2(BaseModel):
    """2 Layer linear-logistic neural network using Keras"""

    parametric = False
    bootlist = None

    def __init__(self, l_lambda=0.01, n_nodes=2, epochs=200, learning_rate=0.01, momentum=0.0, decay=0.0, nesterov=False, loss="binary_crossentropy", batch_size=None, verbose=0):
        self.l_lambda = l_lambda
        self.n_nodes = n_nodes
        self.verbose = verbose
        self.n_epochs = epochs
        self.k = n_nodes
        self.batch_size = batch_size
        self.loss = loss
        self.optimizer = SGD(lr=learning_rate, momentum=momentum, decay=decay, nesterov=nesterov)

    def train(self, X, Y, epoch_ypred=False, epoch_xtest=None):
        """ Fit the neural network model, save additional stats (as attributes) and return Y predicted values.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.

        Y : array-like, shape = [n_samples, 1]
            Response variables, where n_samples is the number of samples.

        Returns
        -------
        y_pred_train : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # If batch-size is None:
        if self.batch_size is None:
            self.batch_size = min(200, len(X))

        self.model = Sequential()
        self.model.add(Dense(self.n_nodes, activation="linear", input_dim=len(X.T), kernel_regularizer=regularizers.l2(self.l_lambda)))
        self.model.add(Dense(1, activation="sigmoid"))
        self.model.compile(optimizer=self.optimizer, loss=self.loss, metrics=["accuracy"])

        # If epoch_ypred is True, calculate ypred for each epoch
        if epoch_ypred is True:
            self.epoch = YpredCallback(self.model, X, epoch_xtest)
        else:
            self.epoch = Callback()

        # Fit
        self.model.fit(X, Y, epochs=self.n_epochs, batch_size=self.batch_size, verbose=self.verbose, callbacks=[self.epoch])

        layer1_weight = self.model.layers[0].get_weights()[0]
        layer1_bias = self.model.layers[0].get_weights()[1]
        layer2_weight = self.model.layers[1].get_weights()[0]
        layer2_bias = self.model.layers[1].get_weights()[1]

        # Not sure about the naming scheme (trying to match PLS)
        self.model.x_loadings_ = layer1_weight
        self.model.x_scores_ = np.matmul(X, self.model.x_loadings_) + layer1_bias
        self.model.y_loadings_ = layer2_weight
        self.model.pctvar_ = np.ones((1, len(self.model.y_loadings_[0])))
        self.xcols_num = len(X.T)
        self.model.pctvar_ = sum(abs(self.model.x_scores_) ** 2) / sum(sum(abs(X) ** 2)) * 100
        y_pred_train = self.model.predict(X).flatten()

        # Storing X, Y, and Y_pred
        self.Y_pred = y_pred_train
        self.X = X
        self.Y = Y
        return y_pred_train

    def test(self, X, Y=None):
        """Calculate and return Y predicted value.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Test variables, where n_samples is the number of samples and n_features is the number of predictors.

        Returns
        -------
        y_pred_test : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        layer1_weight = self.model.layers[0].get_weights()[0]
        layer1_bias = self.model.layers[0].get_weights()[1]
        layer2_weight = self.model.layers[1].get_weights()[0]
        layer2_bias = self.model.layers[1].get_weights()[1]

        self.model.y_loadings_ = layer2_weight
        self.model.y_loadings_ = self.model.y_loadings_.reshape(1, len(self.model.y_loadings_))
        self.model.x_loadings_ = layer1_weight
        self.model.x_scores_ = np.matmul(X, self.model.x_loadings_) + layer1_bias
        y_pred_test = self.model.predict(X).flatten()
        self.model.pctvar_ = sum(abs(self.model.x_scores_) ** 2) / sum(sum(abs(X) ** 2)) * 100
        return y_pred_test



import numpy as np


def dict_median(dict_list):
    median_dict = {}
    for key in dict_list[0].keys():
        value = []
        for i in dict_list:
            value.append(i[key])
        value_arr = np.array(value)
        if np.isnan(value_arr).any() == True:
            median_dict[key] = np.nan
        else:
            median_dict[key] = np.median(value)
    return median_dict



import numpy as np
from keras.callbacks import Callback
from keras.optimizers import SGD
from sklearn.neural_network import MLPClassifier
from keras.models import Sequential
from keras.layers import Dense
from .BaseModel import BaseModel
from ..utils import YpredCallback


class NN_LogitLogit_Sklearn(BaseModel):
    """2 Layer linear-linear neural network using Keras"""

    parametric = False
    bootlist = None

    def __init__(self, n_nodes=2, epochs=200, learning_rate=0.01, momentum=0.0, decay=0.0, nesterov=False, loss="binary_crossentropy", batch_size=None, verbose=0):
        self.n_nodes = n_nodes
        self.verbose = verbose
        self.n_epochs = epochs
        self.k = n_nodes
        self.batch_size = batch_size
        self.loss = loss
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.decay = decay
        self.optimizer = "sgd"

    def train(self, X, Y, epoch_ypred=False, epoch_xtest=None):
        """ Fit the neural network model, save additional stats (as attributes) and return Y predicted values.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.

        Y : array-like, shape = [n_samples, 1]
            Response variables, where n_samples is the number of samples.

        Returns
        -------
        y_pred_train : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # If batch-size is None:
        if self.batch_size is None:
            self.batch_size = len(X)

        # Ensure array and error check
        X, Y = self.input_check(X, Y)

        self.model = MLPClassifier(hidden_layer_sizes=(self.n_nodes,), activation="logistic", solver=self.optimizer, learning_rate_init=self.learning_rate, momentum=self.momentum, batch_size=self.batch_size, nesterovs_momentum=False, max_iter=self.n_epochs)

        # Fit
        self.model.fit(X, Y)

        y_pred_train = self.model.predict(X)

        # Storing X, Y, and Y_pred
        self.Y_pred = y_pred_train
        self.X = X
        self.Y = Y
        return y_pred_train

    def test(self, X, Y=None):
        """Calculate and return Y predicted value.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Test variables, where n_samples is the number of samples and n_features is the number of predictors.

        Returns
        -------
        y_pred_test : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """
        y_pred_test = self.model.predict(X)
        return y_pred_test



import numpy as np


def dict_perc(dict_list, ci=95):
    perc_dict = {}
    for key in dict_list[0].keys():
        value = []
        for i in dict_list:
            value.append(i[key])
        value_arr = np.array(value)
        if np.isnan(value_arr).any() == True:
            perc_dict[key] = [np.nan, np.nan]
        else:
            lower_alpha = (100 - ci) / 2
            upper_alpha = 100 - lower_alpha
            lower_ci = np.percentile(value_arr, lower_alpha)
            upper_ci = np.percentile(value_arr, upper_alpha)
            perc_dict[key] = [lower_ci, upper_ci]
    return perc_dict



import numpy as np
import pandas as pd
import re
import math
import multiprocessing
from abc import ABC, abstractmethod
from sklearn import metrics
from bokeh.layouts import gridplot, layout
from bokeh import events
from bokeh.plotting import figure, output_notebook, show
from bokeh.models import ColumnDataSource, Circle, HoverTool, TapTool, LabelSet, Rect, LinearColorMapper, MultiLine, Patch, Patches, CustomJS, Text, Title
from itertools import product
from sklearn.model_selection import ParameterGrid
from sklearn import preprocessing
from ..utils import color_scale, dict_perc


class BaseCrossVal(ABC):
    """Base class for crossval: kfold."""

    @abstractmethod
    def __init__(self, model, X, Y, param_dict, folds=10, n_mc=1, n_boot=0, n_cores=-1, ci=95):
        # Store basic inputs
        self.model = model
        self.X = X
        self.Y = Y
        self.num_param = len(param_dict)
        self.folds = folds
        self.n_boot = n_boot
        self.n_mc = n_mc
        # Note; if self.mc is 0, change to 1
        if self.n_mc == 0:
            self.n_mc = 1
        self.ci = ci
        # Save param_dict
        # Make sure each parameter is in a list
        for key, value in param_dict.items():
            if not isinstance(value, (list, tuple, np.ndarray)):
                param_dict[key] = [value]
        self.param_dict = param_dict
        self.param_list = list(ParameterGrid(param_dict))
        # Create a second dict, with parameters with more than 1 variable e.g. n_comp = [1, 2]
        self.param_dict2 = {}
        for key, value in param_dict.items():
            if len(value) > 1:
                self.param_dict2 = {**self.param_dict2, **{key: value}}
        self.param_list2 = list(ParameterGrid(self.param_dict2))

        # if n_cores = -1, set n_cores to max_cores
        max_num_cores = multiprocessing.cpu_count()
        self.n_cores = n_cores
        if self.n_cores > max_num_cores:
            self.n_cores = -1
            print("Number of cores set too high. It will be set to the max number of cores in the system.", flush=True)
        if self.n_cores == -1:
            self.n_cores = max_num_cores
            print("Number of cores set to: {}".format(max_num_cores))

    @abstractmethod
    def calc_ypred(self):
        """Calculates ypred full and ypred cv."""
        pass

    @abstractmethod
    def calc_ypred_epoch(self):
        """Calculates ypred full and ypred cv."""
        pass

    @abstractmethod
    def calc_stats(self):
        """Calculates binary statistics from ypred full and ypred cv."""
        pass

    def _format_table(self, stats_list):
        """Make stats pretty (pandas table -> proper names in columns)."""
        table = pd.DataFrame(stats_list).T
        param_list_string = []
        for i in range(len(self.param_list)):
            param_list_string.append(str(self.param_list[i]))
        table.columns = param_list_string
        return table

    def run(self):
        """Runs all functions prior to plot."""
        # Check that param_dict is not for epochs
        # Epoch is a special case
        print("Running ...")
        check_epoch = []
        for i in self.param_dict2.keys():
            check_epoch.append(i)
        if check_epoch == ["epochs"]:
            # Get epoch max
            epoch_list = []
            for i in self.param_list2:
                for k, v in i.items():
                    epoch_list.append(v)
            # Print and Calculate
            self.calc_ypred_epoch()
            print("returning stats at 'x' epoch interval during training until epoch={}.".format(epoch_list[-1]))
        else:
            self.calc_ypred()
        self.calc_stats()
        print("Done!")

    def plot(self, metric="r2q2", scale=1, color_scaling="tanh", rotate_xlabel=True, model="kfold", legend="bottom_right", color_beta=[10, 10, 10], ci=95):
        """Create a full/cv plot using based on metric selected.

        Parameters
        ----------
        metric : string, (default "r2q2")
            metric has to be either "r2q2", "auc", "acc", "f1score", "prec", "sens", or "spec".
        """
        # Check model is parametric if using 'r2q2'
        if metric == "r2q2":
            if self.model.parametric is False:
                print("metric changed from 'r2q2' to 'auc' as the model is non-parametric.")
                metric = "auc"

        # Plot based on the number of parameters
        if len(self.param_dict2) == 1:
            fig = self._plot_param1(metric=metric, scale=scale, rotate_xlabel=rotate_xlabel, model=model, legend=legend, ci=ci)
        elif len(self.param_dict2) == 2:
            fig = self._plot_param2(metric=metric, scale=scale, color_scaling=color_scaling, model=model, legend=legend, color_beta=color_beta, ci=ci)
        else:
            raise ValueError("plot function only works for 1 or 2 parameters, there are {}.".format(len(self.param_dict2)))

        # Show plot
        output_notebook()
        show(fig)

    def _plot_param1(self, metric="r2q2", scale=1, rotate_xlabel=True, model="kfold", title_align="center", legend="bottom_right", ci=95):
        """Used for plot function if the number of parameters is 1."""

        # Get ci
        if self.n_mc > 1:
            std_list = []
            for i in range(len(self.param_list)):
                std_full_i = dict_perc(self.full_loop[i], ci=ci)
                std_cv_i = dict_perc(self.cv_loop[i], ci=ci)
                std_full_i = {k + "full": v for k, v in std_full_i.items()}
                std_cv_i = {k + "cv": v for k, v in std_cv_i.items()}
                std_cv_i["R²"] = std_full_i.pop("R²full")
                std_cv_i["Q²"] = std_cv_i.pop("R²cv")
                std_combined = {**std_full_i, **std_cv_i}
                std_list.append(std_combined)
            self.table_std = self._format_table(std_list)  # Transpose, Add headers

        # Choose metric to plot
        metric_title = np.array(["ACCURACY", "AIC", "AUC", "BIC", "F1-SCORE", "PRECISION", "R²", "SENSITIVITY", "SPECIFICITY", "SSE"])
        metric_list = np.array(["acc", "aic", "auc", "bic", "f1score", "prec", "r2q2", "sens", "spec", "sse"])
        metric_idx = np.where(metric_list == metric)[0][0]

        # get full, cv, and diff
        full = self.table.iloc[2 * metric_idx + 1]
        cv = self.table.iloc[2 * metric_idx]
        diff = abs(full - cv)
        full_text = self.table.iloc[2 * metric_idx + 1].name
        cv_text = self.table.iloc[2 * metric_idx].name
        if metric == "r2q2":
            diff_text = "| R²-Q² |"
            y_axis_text = "R² & Q²"
            full_legend = "R²"
            cv_legend = "Q²"
        else:
            diff_text = full_text[:-4] + "diff"
            y_axis_text = full_text[:-4]
            if model == "kfold":
                full_legend = "FULL"
                cv_legend = "CV"
            else:
                full_legend = "TRAIN"
                cv_legend = "TEST"
                full_text = full_text[:-4] + "train"
                cv_text = full_text[:-5] + "test"

        # round full, cv, and diff for hovertool
        full_hover = []
        cv_hover = []
        diff_hover = []
        for j in range(len(full)):
            full_hover.append("%.2f" % round(full[j], 2))
            cv_hover.append("%.2f" % round(cv[j], 2))
            diff_hover.append("%.2f" % round(diff[j], 2))

        # get key, values (as string) from param_dict (key -> title, values -> x axis values)
        for k, v in self.param_dict2.items():
            key_title = k
            key_xaxis = k
            values = v
        values_string = [str(i) for i in values]
        values_string = []
        for i in values:
            if i == 0:
                values_string.append(str(i))
            elif 0.0001 > i:
                values_string.append("%0.2e" % i)
            elif 10000 < i:
                values_string.append("%0.2e" % i)
            else:
                values_string.append(str(i))

        # if parameter starts with n_ e.g. n_components change title to 'no. of components', xaxis to 'components'
        if key_title.startswith("n_"):
            key_xaxis = key_xaxis[2:]
            key_xaxis = key_xaxis.title()
            key_title = "no. of " + key_xaxis
        else:
            key_title = key_title.replace("_", " ")
            key_title = key_title.title()
            key_xaxis = key_title
            # if key_xaxis.endswith("s") == True:
            #     key_xaxis = key_xaxis[:-1]

        # store data in ColumnDataSource for Bokeh
        data = dict(full=full, cv=cv, diff=diff, full_hover=full_hover, cv_hover=cv_hover, diff_hover=diff_hover, values_string=values_string)
        source = ColumnDataSource(data=data)

        # fig1_yrange = (min(diff) - max(0.1 * (min(diff)), 0.07), max(diff) + max(0.1 * (max(diff)), 0.07))
        # fig1_xrange = (min(cv) - max(0.1 * (min(cv)), 0.07), max(cv) + max(0.1 * (max(cv)), 0.07))
        fig1_title = diff_text + " vs. " + cv_text

        # Plot width/height
        width = int(485 * scale)
        height = int(405 * scale)
        # fig1_yrange = (min(diff) - max(0.1 * (min(diff)), 0.07), max(diff) + max(0.1 * (max(diff)), 0.07))
        # fig1_xrange = (min(cv) - max(0.1 * (min(cv)), 0.07), max(cv) + max(0.1 * (max(cv)), 0.07))
        # x_range=(min(cv_score) - 0.03, max(cv_score) + 0.03), y_range=(min(diff_score) - 0.03, max(diff_score) + 0.03)

        # Figure 1 (DIFFERENCE (R2 - Q2) vs. Q2)
        fig1 = figure(x_axis_label=cv_text, y_axis_label=diff_text, title=fig1_title, tools="tap,pan,wheel_zoom,box_zoom,reset,save,lasso_select,box_select", plot_width=width, plot_height=height, x_range=(min(cv) - 0.03, max(cv) + 0.03), y_range=(min(diff) - 0.03, max(diff) + 0.03))

        # Figure 1: Add a line
        fig1_line = fig1.line(cv, diff, line_width=2, line_color="black", line_alpha=0.25)

        # Figure 1: Add circles (interactive click)
        fig1_circ = fig1.circle("cv", "diff", size=12, alpha=0.7, color="green", source=source)
        fig1_circ.selection_glyph = Circle(fill_color="green", line_width=2, line_color="black")
        fig1_circ.nonselection_glyph.fill_color = "green"
        fig1_circ.nonselection_glyph.fill_alpha = 0.4
        fig1_circ.nonselection_glyph.line_color = "white"

        # Figure 1: Add hovertool
        fig1.add_tools(HoverTool(renderers=[fig1_circ], tooltips=[(key_xaxis, "@values_string"), (full_text, "@full_hover"), (cv_text, "@cv_hover"), (diff_text, "@diff_hover")]))

        # Figure 1: Extra formating
        fig1.axis.major_label_text_font_size = "8pt"
        if metric is "r2q2" or metric is "auc":
            fig1.title.text_font_size = "12pt"
            fig1.xaxis.axis_label_text_font_size = "10pt"
            fig1.yaxis.axis_label_text_font_size = "10pt"
        else:
            fig1.title.text_font_size = "10pt"
            fig1.xaxis.axis_label_text_font_size = "9pt"
            fig1.yaxis.axis_label_text_font_size = "9pt"

        # Figure 2: full/cv
        fig2_title = y_axis_text + " over " + key_title
        fig2 = figure(x_axis_label=key_xaxis, y_axis_label=y_axis_text, title=fig2_title, plot_width=width, plot_height=height, x_range=pd.unique(values_string), tools="pan,wheel_zoom,box_zoom,reset,save,lasso_select,box_select")

        # Figure 2: Add Confidence Intervals if n_mc > 1
        if self.n_mc > 1:
            # get full, cv, and diff
            full_std = self.table_std.iloc[2 * metric_idx + 1]
            cv_std = self.table_std.iloc[2 * metric_idx]
            lower_ci_full = pd.Series(name=full_std.name, dtype="object")
            upper_ci_full = pd.Series(name=full_std.name, dtype="object")
            for key, values in full_std.iteritems():
                lower_ci_full[key] = values[0]
                upper_ci_full[key] = values[1]
            lower_ci_cv = pd.Series(name=cv_std.name, dtype="object")
            upper_ci_cv = pd.Series(name=cv_std.name, dtype="object")
            for key, values in cv_std.iteritems():
                lower_ci_cv[key] = values[0]
                upper_ci_cv[key] = values[1]
            # Plot as a patch
            x_patch = np.hstack((values_string, values_string[::-1]))
            y_patch_r2 = np.hstack((lower_ci_full, upper_ci_full[::-1]))
            y_patch_q2 = np.hstack((lower_ci_cv, upper_ci_cv[::-1]))
            fig2.patch(x_patch, y_patch_q2, alpha=0.10, color="blue")
            # kfold monte-carlo does not have ci for R2
            if model is not "kfold":
                fig2.patch(x_patch, y_patch_r2, alpha=0.10, color="red")

        # Figure 2: add full
        fig2_line_full = fig2.line(values_string, full, line_color="red", line_width=2)
        fig2_circ_full = fig2.circle("values_string", "full", line_color="red", fill_color="white", fill_alpha=1, size=8, source=source, legend=full_legend)
        fig2_circ_full.selection_glyph = Circle(line_color="red", fill_color="white", line_width=2)
        fig2_circ_full.nonselection_glyph.line_color = "red"
        fig2_circ_full.nonselection_glyph.fill_color = "white"
        fig2_circ_full.nonselection_glyph.line_alpha = 0.4

        # Figure 2: add cv
        fig2_line_cv = fig2.line(values_string, cv, line_color="blue", line_width=2)
        fig2_circ_cv = fig2.circle("values_string", "cv", line_color="blue", fill_color="white", fill_alpha=1, size=8, source=source, legend=cv_legend)
        fig2_circ_cv.selection_glyph = Circle(line_color="blue", fill_color="white", line_width=2)
        fig2_circ_cv.nonselection_glyph.line_color = "blue"
        fig2_circ_cv.nonselection_glyph.fill_color = "white"
        fig2_circ_cv.nonselection_glyph.line_alpha = 0.4

        # Add hovertool and taptool
        fig2.add_tools(HoverTool(renderers=[fig2_circ_full], tooltips=[(full_text, "@full_hover")], mode="vline"))
        fig2.add_tools(HoverTool(renderers=[fig2_circ_cv], tooltips=[(cv_text, "@cv_hover")], mode="vline"))
        fig2.add_tools(TapTool(renderers=[fig2_circ_full, fig2_circ_cv]))

        # Figure 2: Extra formating
        fig2.axis.major_label_text_font_size = "8pt"
        if metric is "r2q2" or metric is "auc":
            fig2.title.text_font_size = "12pt"
            fig2.xaxis.axis_label_text_font_size = "10pt"
            fig2.yaxis.axis_label_text_font_size = "10pt"
        else:
            fig2.title.text_font_size = "10pt"
            fig2.xaxis.axis_label_text_font_size = "9pt"
            fig2.yaxis.axis_label_text_font_size = "9pt"

        # Rotate
        if rotate_xlabel is True:
            fig2.xaxis.major_label_orientation = np.pi / 2

        # Figure 2: legend
        if legend == None or legend == False:
            fig2.legend.visible = False
        else:
            fig2.legend.location = legend
            fig2.legend.location = legend

        # Hide legend if it is clicked
        # def show_hide_legend(legend=fig2.legend[0]):
        #     legend.visible = not legend.visible

        # print(py2js(show_hide_legend))

        # fig2.js_on_event(events.DoubleTap, CustomJS.from_py_func(show_hide_legend))

        # Center title
        if title_align == "center":
            fig1.title.align = "center"
            fig2.title.align = "center"

        # Create a grid and output figures
        grid = np.full((1, 2), None)
        grid[0, 0] = fig1
        grid[0, 1] = fig2
        fig = gridplot(grid.tolist(), merge_tools=True)
        return fig

    def _plot_param2(self, metric="r2q2", xlabel=None, orientation=0, alternative=False, scale=1, heatmap_xaxis_rotate=90, color_scaling="tanh", line=False, model="kfold", title_align="center", legend="bottom_right", color_beta=[10, 10, 10], ci=95):

        # legend always None
        legend = None

        # check color_beta
        if type(color_beta) != list:
            raise ValueError("color_beta needs to be a list of 3 values e.g. [10, 10, 10]")
        if len(color_beta) != 3:
            raise ValueError("color_beta needs to be a list of 3 values e.g. [10, 10, 10]")

        # Get ci
        if self.n_mc > 1:
            std_list = []
            for i in range(len(self.param_list)):
                std_full_i = dict_perc(self.full_loop[i], ci=ci)
                std_cv_i = dict_perc(self.cv_loop[i], ci=ci)
                std_full_i = {k + "full": v for k, v in std_full_i.items()}
                std_cv_i = {k + "cv": v for k, v in std_cv_i.items()}
                std_cv_i["R²"] = std_full_i.pop("R²full")
                std_cv_i["Q²"] = std_cv_i.pop("R²cv")
                std_combined = {**std_full_i, **std_cv_i}
                std_list.append(std_combined)
            self.table_std = self._format_table(std_list)  # Transpose, Add headers

        metric_list = np.array(["acc", "aic", "auc", "bic", "f1score", "prec", "r2q2", "sens", "spec", "sse"])
        metric_idx = np.where(metric_list == metric)[0][0]

        # get full, cv, and diff
        full_score = self.table.iloc[2 * metric_idx + 1]
        cv_score = self.table.iloc[2 * metric_idx]
        diff_score = abs(full_score - cv_score)
        full_title = self.table.iloc[2 * metric_idx + 1].name
        cv_title = self.table.iloc[2 * metric_idx].name
        diff_title = full_title[:-4] + "diff"

        y_axis_text = full_title[:-4]
        if model == "kfold":
            full_legend = "FULL"
            cv_legend = "CV"
        else:
            full_legend = "TRAIN"
            cv_legend = "TEST"
            full_title = full_title[:-4] + "train"
            cv_title = full_title[:-5] + "test"

        # round full, cv, and diff for hovertool
        full_hover = []
        cv_hover = []
        diff_hover = []
        for j in range(len(full_score)):
            full_hover.append("%.2f" % round(full_score[j], 2))
            cv_hover.append("%.2f" % round(cv_score[j], 2))
            diff_hover.append("%.2f" % round(diff_score[j], 2))

        # If n_mc > 1
        if self.n_mc > 1:
            # get full, cv, and diff
            full_std = self.table_std.iloc[2 * metric_idx + 1]
            cv_std = self.table_std.iloc[2 * metric_idx]
            lower_ci_full = pd.Series(name=full_std.name, dtype="object")
            upper_ci_full = pd.Series(name=full_std.name, dtype="object")
            for key, values in full_std.iteritems():
                lower_ci_full[key] = values[0]
                upper_ci_full[key] = values[1]
            lower_ci_cv = pd.Series(name=cv_std.name, dtype="object")
            upper_ci_cv = pd.Series(name=cv_std.name, dtype="object")
            for key, values in cv_std.iteritems():
                lower_ci_cv[key] = values[0]
                upper_ci_cv[key] = values[1]

        # Get key/values
        param_keys = []
        param_values = []
        for key, value in sorted(self.param_dict2.items()):
            param_keys.append(key)
            # value_to_string = list(map(str, value))
            # param_values.append(value_to_string)
            values_string = []
            for i in value:
                if i == 0:
                    values_string.append(str(i))
                elif 0.0001 > i:
                    values_string.append("%0.2e" % i)
                elif 10000 < i:
                    values_string.append("%0.2e" % i)
                else:
                    values_string.append(str(i))

            param_values.append(values_string)

        param_keys_title = []
        param_keys_axis = []
        for i in param_keys:
            if i.startswith("n_"):
                i_xaxis = i[2:]
                i_xaxis = i_xaxis.title()
                i_title = "no. of " + i_xaxis
            else:
                i_title = i.replace("_", " ")
                i_title = i_title.title()
                i_xaxis = i_title
            param_keys_title.append(i_title)
            param_keys_axis.append(i_xaxis)

        # Get key/value combinations
        comb = list(product(param_values[0], param_values[1]))
        key0_value = [val[0] for val in comb]
        key1_value = [val[1] for val in comb]
        key0_unique = param_values[0]
        key1_unique = param_values[1]
        table = self.table
        param_dict = self.param_dict2
        param_list = self.param_list2

        # diff_alpha_input = diff_score.copy()
        # for i in diff_alpha_input.index:
        #     diff_alpha_input[i] = -diff_alpha_input[i]

        full_alpha = color_scale(full_score, method=color_scaling, beta=color_beta[0])
        cv_alpha = color_scale(cv_score, method=color_scaling, beta=color_beta[1])
        diff_alpha = color_scale(diff_score, method=color_scaling, beta=color_beta[2])
        # diff_alpha = 1.1 - diff_alpha

        # Text for heatmaps
        full_text = []
        cv_text = []
        diff_text = []
        for i in range(len(key0_value)):
            full_text.append("%.2f" % round(full_score[i], 2))
            cv_text.append("%.2f" % round(cv_score[i], 2))
            diff_text.append("%.2f" % round(diff_score[i], 2))

        # Information for line plot
        line_key0_value = []
        for i in range(len(key0_value)):
            line_key0_value.append(key0_unique)
        line_key1_value = []
        for i in range(len(key1_value)):
            line_key1_value.append(key1_unique)

        # Extra for n_mc
        if self.n_mc > 1:
            # Information for line plot
            monte_line_key0_value = []
            for i in range(len(key0_value)):
                monte_line_key0_value.append(list(np.hstack((key0_unique, key0_unique[::-1]))))
            monte_line_key1_value = []
            for i in range(len(key1_value)):
                monte_line_key1_value.append(list(np.hstack((key1_unique, key1_unique[::-1]))))

        line0_full = []
        line0_cv = []
        for i in range(len(key0_value)):
            line0_full_i = []
            line0_cv_i = []
            for j in range(len(key0_value)):
                if key0_value[i] == key0_value[j]:
                    line0_full_i.append(full_score[j])
                    line0_cv_i.append(cv_score[j])
            line0_full.append(line0_full_i)
            line0_cv.append(line0_cv_i)

        line1_full = []
        line1_cv = []
        for i in range(len(key1_value)):
            line1_full_i = []
            line1_cv_i = []
            for j in range(len(key1_value)):
                if key1_value[i] == key1_value[j]:
                    line1_full_i.append(full_score[j])
                    line1_cv_i.append(cv_score[j])
            line1_full.append(line1_full_i)
            line1_cv.append(line1_cv_i)

        # Extra for n_mc
        if self.n_mc > 1:
            monte_line1_full = []
            monte_line1_cv = []
            for i in range(len(key1_value)):
                line1_full_i_upper = []
                line1_full_i_lower = []
                line1_cv_i_upper = []
                line1_cv_i_lower = []
                for j in range(len(key1_value)):
                    if key1_value[i] == key1_value[j]:
                        line1_full_i_upper.append(upper_ci_full[j])
                        line1_full_i_lower.append(lower_ci_full[j])
                        line1_cv_i_upper.append(upper_ci_cv[j])
                        line1_cv_i_lower.append(lower_ci_cv[j])
                monte_line1_full.append(list(np.hstack((line1_full_i_lower, line1_full_i_upper[::-1]))))
                monte_line1_cv.append(list(np.hstack((line1_cv_i_lower, line1_cv_i_upper[::-1]))))

        # Extra for n_mc
        if self.n_mc > 1:
            monte_line0_full = []
            monte_line0_cv = []
            for i in range(len(key0_value)):
                line0_full_i_upper = []
                line0_full_i_lower = []
                line0_cv_i_upper = []
                line0_cv_i_lower = []
                for j in range(len(key0_value)):
                    if key0_value[i] == key0_value[j]:
                        line0_full_i_upper.append(upper_ci_full[j])
                        line0_full_i_lower.append(lower_ci_full[j])
                        line0_cv_i_upper.append(upper_ci_cv[j])
                        line0_cv_i_lower.append(lower_ci_cv[j])
                monte_line0_full.append(list(np.hstack((line0_full_i_lower, line0_full_i_upper[::-1]))))
                monte_line0_cv.append(list(np.hstack((line0_cv_i_lower, line0_cv_i_upper[::-1]))))

        # Scatterplot color and size based on key0 and key1
        color_key0 = []
        for i in range(len(key0_value)):
            for j in range(len(key0_unique)):
                if key0_value[i] == key0_unique[j]:
                    color_key0.append(j / (len(key0_unique) - 1))

        scaler_size = preprocessing.MinMaxScaler(feature_range=(4, 20))
        size_prescale_key1 = []

        for i in range(len(key1_value)):
            for j in range(len(key1_unique)):
                if key1_value[i] == key1_unique[j]:
                    size_prescale_key1.append(j / (len(key1_unique) - 1))
        scatter_size_key1 = scaler_size.fit_transform(np.array(size_prescale_key1)[:, np.newaxis])
        scatter_size_key1 = scatter_size_key1 * scale

        # Extra
        key0_value_text = len(key1_value) * [key1_value[-1]]
        key1_value_text = len(key0_value) * [key0_value[-1]]
        line1_cv_text = []
        for i in line1_cv:
            line1_cv_text.append(i[-1])
        line0_cv_text = []
        for i in line0_cv:
            line0_cv_text.append(i[-1])
        line1_full_text = []
        for i in line1_full:
            line1_full_text.append(i[-1])
        line0_full_text = []
        for i in line0_full:
            line0_full_text.append(i[-1])

        ptext_is = ["Learning Rate", "Nodes", "Neurons", "Momentum", "Decay", "Components", "Batch Size", "Gamma", "C", "Estimators", "Max Features", "Max Depth", "Min Samples Split", "Min Samples Leaf", "Max Leaf Nodes"]
        ptext_change = ["LR", "Node", "Neur", "Mom", "Dec", "Comp", "Bat", "Gam", "C", "Est", "Feat", "Dep", "SSpl", "SLea", "LNod"]

        ptext = []
        for i in param_keys_axis:
            val = "fill"
            for j in range(len(ptext_is)):
                if i == ptext_is[j]:
                    val = ptext_change[j]
            if val == "fill":
                val = i[:3]
            ptext.append(val + " = ")

        line1_cv_score_text = []
        for i in key1_value:
            line1_cv_score_text.append(ptext[1] + i)

        line0_cv_score_text = []
        for i in key0_value:
            line0_cv_score_text.append(ptext[0] + i)

        diff_score_neg = 1 - diff_score
        # Store information in dictionary for bokeh
        data = dict(
            key0_value=key1_value,
            key1_value=key0_value,
            full_score=full_score,
            cv_score=cv_score,
            diff_score=diff_score,
            diff_score_neg=diff_score_neg,
            full_alpha=full_alpha,
            cv_alpha=cv_alpha,
            diff_alpha=diff_alpha,
            line_key0_value=line_key0_value,
            line_key1_value=line_key1_value,
            line0_full=line0_full,
            line0_cv=line0_cv,
            line1_full=line1_full,
            line1_cv=line1_cv,
            full_text=full_text,
            cv_text=cv_text,
            diff_text=diff_text,
            key0_value_text=key0_value_text,
            key1_value_text=key1_value_text,
            line0_cv_text=line0_cv_text,
            line1_cv_text=line1_cv_text,
            line1_cv_score_text=line1_cv_score_text,
            line1_full_text=line1_full_text,
            line0_full_text=line0_full_text,
            line0_cv_score_text=line0_cv_score_text,
        )

        if self.n_mc > 1:
            data["lower_ci_full"] = lower_ci_full
            data["upper_ci_full"] = upper_ci_full
            data["lower_ci_cv"] = lower_ci_cv
            data["upper_ci_cv"] = lower_ci_cv
            data["monte_line_key1_value"] = monte_line_key1_value
            data["monte_line_key0_value"] = monte_line_key0_value
            data["monte_line1_full"] = monte_line1_full
            data["monte_line1_cv"] = monte_line1_cv
            data["monte_line0_full"] = monte_line0_full
            data["monte_line0_cv"] = monte_line0_cv
        source = ColumnDataSource(data=data)

        # Heatmap FULL
        p1 = figure(title=full_title, tools="tap, save", x_range=key0_unique, y_range=key1_unique, x_axis_label=param_keys_axis[0], y_axis_label=param_keys_axis[1])

        p1_render = p1.rect("key1_value", "key0_value", 0.9, 0.9, color="red", alpha="full_alpha", line_color=None, source=source)

        p1_render.selection_glyph = Rect(fill_color="red", fill_alpha="full_alpha", line_width=int(3 * scale), line_color="black")
        p1_render.nonselection_glyph.fill_alpha = "full_alpha"
        p1_render.nonselection_glyph.fill_color = "red"
        p1_render.nonselection_glyph.line_color = "white"

        # Heatmap CV
        p2 = figure(title=cv_title, tools="tap, save", x_range=key0_unique, y_range=key1_unique, x_axis_label=param_keys_axis[0], y_axis_label=param_keys_axis[1])

        p2_render = p2.rect("key1_value", "key0_value", 0.9, 0.9, color="blue", alpha="cv_alpha", line_color=None, source=source)

        p2_render.selection_glyph = Rect(fill_color="blue", fill_alpha="cv_alpha", line_width=int(3 * scale), line_color="black")
        p2_render.nonselection_glyph.fill_alpha = "cv_alpha"
        p2_render.nonselection_glyph.fill_color = "blue"
        p2_render.nonselection_glyph.line_color = "white"

        # Heatmap Diff
        p3 = figure(title=diff_title, tools="tap, save", x_range=key0_unique, y_range=key1_unique, x_axis_label=param_keys_axis[0], y_axis_label=param_keys_axis[1])

        p3_render = p3.rect("key1_value", "key0_value", 0.9, 0.9, color="green", alpha="diff_alpha", line_color=None, source=source)

        p3_render.selection_glyph = Rect(fill_color="green", fill_alpha="diff_alpha", line_width=int(3 * scale), line_color="black")
        p3_render.nonselection_glyph.fill_alpha = "diff_alpha"
        p3_render.nonselection_glyph.fill_color = "green"
        p3_render.nonselection_glyph.line_color = "white"

        # Extra for heatmaps
        p1.plot_width = int(320 * scale)
        p1.plot_height = int(257 * scale)
        p1.grid.grid_line_color = None
        p1.axis.axis_line_color = None
        p1.axis.major_tick_line_color = None
        p1.axis.major_label_text_font_size = str(8 * scale) + "pt"
        p1.axis.major_label_standoff = 0
        p1.xaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p1.yaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p1.title.text_font_size = str(14 * scale) + "pt"
        p1.xaxis.major_label_orientation = math.radians(heatmap_xaxis_rotate)

        p2.plot_width = int(320 * scale)
        p2.plot_height = int(257 * scale)
        p2.grid.grid_line_color = None
        p2.axis.axis_line_color = None
        p2.axis.major_tick_line_color = None
        p2.axis.major_label_text_font_size = str(8 * scale) + "pt"
        p2.axis.major_label_standoff = 0
        p2.xaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p2.yaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p2.title.text_font_size = str(14 * scale) + "pt"
        p2.xaxis.major_label_orientation = math.radians(heatmap_xaxis_rotate)

        p3.plot_width = int(320 * scale)
        p3.plot_height = int(257 * scale)
        p3.grid.grid_line_color = None
        p3.axis.axis_line_color = None
        p3.axis.major_tick_line_color = None
        p3.axis.major_label_text_font_size = str(8 * scale) + "pt"
        p3.axis.major_label_standoff = 0
        p3.xaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p3.yaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p3.title.text_font_size = str(14 * scale) + "pt"
        p3.xaxis.major_label_orientation = math.radians(heatmap_xaxis_rotate)

        text = False
        # Adding text to heatmaps
        if text is True:
            # if heatmap rect is dark, use light text and vise versa
            color_mapper_diff = LinearColorMapper(palette=["#000000", "#010101", "#fdfdfd", "#fefefe", "#ffffff"], low=0, high=1)

            label1 = LabelSet(x="key0_value", y="key1_value", text="full_text", level="glyph", x_offset=-10 * scale, y_offset=-10 * scale, source=source, render_mode="canvas", text_font_size=str(7.5 * scale) + "pt", text_color={"field": "full_alpha", "transform": color_mapper_diff})

            label2 = LabelSet(x="key0_value", y="key1_value", text="cv_text", level="glyph", x_offset=-10 * scale, y_offset=-10 * scale, source=source, render_mode="canvas", text_font_size=str(7.5 * scale) + "pt", text_color={"field": "cv_alpha", "transform": color_mapper_diff})

            label3 = LabelSet(x="key0_value", y="key1_value", text="diff_text", level="glyph", x_offset=-10 * scale, y_offset=-10 * scale, source=source, render_mode="canvas", text_font_size=str(7.5 * scale) + "pt", text_color={"field": "diff_alpha", "transform": color_mapper_diff})

            p1.add_layout(label1)
            p2.add_layout(label2)
            p3.add_layout(label3)

        p1.add_tools(HoverTool(renderers=[p1_render, p2_render, p3_render], tooltips=[("AUC_full", "@full_text")]))
        p2.add_tools(HoverTool(renderers=[p1_render, p2_render, p3_render], tooltips=[("AUC_CV", "@cv_text")]))
        p3.add_tools(HoverTool(renderers=[p1_render, p2_render, p3_render], tooltips=[("AUC_diff", "@diff_text")]))

        sc_title = diff_title + " vs. " + cv_title
        # Scatterplot
        p4 = figure(title=sc_title, x_axis_label=cv_title, y_axis_label=diff_title, tools="tap,pan,wheel_zoom,box_zoom,reset,save,lasso_select,box_select", x_range=(min(cv_score) - 0.03, max(cv_score) + 0.03), y_range=(min(diff_score) - 0.03, max(diff_score) + 0.03))

        color_mapper_scatter = LinearColorMapper(palette="Inferno256", low=0, high=1)

        p4_render = p4.circle("cv_score", "diff_score", size=8 * scale, alpha=0.6, color="green", source=source)
        p4_render.selection_glyph = Circle(fill_color="green", line_width=int(2 * scale), line_color="black")
        p4_render.nonselection_glyph.fill_color = "green"
        p4_render.nonselection_glyph.fill_alpha = 0.4
        p4_render.nonselection_glyph.line_color = "white"
        p4.add_tools(HoverTool(renderers=[p4_render], tooltips=[("AUC_full", "@full_text"), ("AUC_CV", "@cv_text"), ("AUC_diff", "@diff_text")]))

        p4.plot_width = int(320 * scale)
        p4.plot_height = int(257 * scale)
        p4.axis.major_label_text_font_size = str(8 * scale) + "pt"
        p4.xaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p4.yaxis.axis_label_text_font_size = str(12 * scale) + "pt"
        p4.title.text_font_size = str(14 * scale) + "pt"

        # Line plot 1
        l1_range_special = []
        if len(key0_unique) > 2:
            l1_range_special.append([" "])
        if len(key0_unique) > 5:
            l1_range_special.append([l1_range_special[-1][0] + " "])
        if len(key0_unique) > 8:
            l1_range_special.append([l1_range_special[-1][0] + " "])
        another_val = len(key0_unique) - 8
        if another_val > 0:
            for i in range(another_val):
                if i % 3 == 0:
                    l1_range_special.append([l1_range_special[-1][0] + " "])
        l1_xrange = pd.unique(key0_unique)
        l1_xrange2 = np.append(l1_xrange, l1_range_special)
        l1_title = y_axis_text + " over " + param_keys_title[0]

        y_range_min = min(cv_score) - min(cv_score) * 0.1
        y_range_max = max(full_score) + max(full_score) * 0.05

        p5 = figure(title=l1_title, x_axis_label=param_keys_axis[0], y_axis_label=y_axis_text, plot_width=int(320 * scale), plot_height=int(257 * scale), x_range=l1_xrange2, tools="pan,wheel_zoom,box_zoom,reset,save,lasso_select,box_select", y_range=(y_range_min, y_range_max))
        p5.quad(top=[1000], bottom=[-1000], left=[l1_xrange[-1]], right=[1000], color="white")

        # p5.outline_line_color = "white"
        if self.n_mc > 1:
            p5_render_patch2 = p5.patches("monte_line_key0_value", "monte_line1_cv", alpha=0, color="blue", source=source)
            p5_render_patch2.selection_glyph = Patches(fill_alpha=0.2, fill_color="blue", line_color="white")
            p5_render_patch2.nonselection_glyph.fill_alpha = 0
            p5_render_patch2.nonselection_glyph.line_color = "white"
            # kfold monte-carlo does not have ci for R2
            if model is not "kfold":
                p5_render_patch1 = p5.patches("monte_line_key0_value", "monte_line1_full", alpha=0, color="red", source=source)
                p5_render_patch1.selection_glyph = Patches(fill_alpha=0.2, fill_color="red", line_color="white")
                p5_render_patch1.nonselection_glyph.fill_alpha = 0
                p5_render_patch1.nonselection_glyph.line_color = "white"

        p5_render_1 = p5.multi_line("line_key0_value", "line1_full", line_color="red", line_width=2 * scale, source=source)
        p5_render_1.selection_glyph = MultiLine(line_color="red", line_alpha=0.8, line_width=2 * scale)
        p5_render_1.nonselection_glyph.line_color = "red"
        p5_render_1.nonselection_glyph.line_alpha = 0.05 / len(key1_unique)

        p5_render_2 = p5.circle("key1_value", "full_score", line_color="red", fill_color="white", size=8 * scale, source=source, legend=full_legend)
        p5_render_2.selection_glyph = Circle(line_color="red", fill_color="white")
        p5_render_2.nonselection_glyph.line_color = "red"
        p5_render_2.nonselection_glyph.fill_color = "white"
        p5_render_2.nonselection_glyph.line_alpha = 0.7 / len(key1_unique)

        p5_render_3 = p5.multi_line("line_key0_value", "line1_cv", line_color="blue", line_width=2 * scale, source=source)
        p5_render_3.selection_glyph = MultiLine(line_color="blue", line_alpha=0.8, line_width=2 * scale)
        p5_render_3.nonselection_glyph.line_color = "blue"
        p5_render_3.nonselection_glyph.line_alpha = 0.05 / len(key1_unique)

        p5_render_4 = p5.circle("key1_value", "cv_score", line_color="blue", fill_color="white", size=8 * scale, source=source, legend=cv_legend)
        p5_render_4.selection_glyph = Circle(line_color="blue", fill_color="white")
        p5_render_4.nonselection_glyph.line_color = "blue"
        p5_render_4.nonselection_glyph.fill_color = "white"
        p5_render_4.nonselection_glyph.line_alpha = 0.7 / len(key1_unique)

        # text
        text_here = 8 * scale
        text_line_font = str(text_here) + "pt"
        p5_render_5 = p5.text(x="key1_value_text", y="line1_cv_text", text="line1_cv_score_text", source=source, text_font_size=text_line_font, text_color="blue", x_offset=8 * scale, y_offset=6 * scale, text_alpha=0)
        p5_render_5.selection_glyph = Text(text_color="blue", text_alpha=1, text_font_size=text_line_font)
        p5_render_5.nonselection_glyph.text_alpha = 0
        p5_render_6 = p5.text(x="key1_value_text", y="line1_full_text", text="line1_cv_score_text", source=source, text_font_size=text_line_font, text_color="red", x_offset=8 * scale, y_offset=6 * scale, text_alpha=0)
        p5_render_6.selection_glyph = Text(text_color="red", text_alpha=1, text_font_size=text_line_font)
        p5_render_6.nonselection_glyph.text_alpha = 0

        # p5_render_5.selection_glyph.text_alpha = 0.6

        p5.add_tools(HoverTool(renderers=[p5_render_2], tooltips=[("AUC_full", "@full_text")]))

        p5.add_tools(HoverTool(renderers=[p5_render_4], tooltips=[("AUC_CV", "@cv_text")]))

        p5.add_tools(TapTool(renderers=[p5_render_2, p5_render_4]))

        # Line plot 2
        l2_range_special = []
        if len(key1_unique) > 2:
            l2_range_special.append([" "])
        if len(key1_unique) > 5:
            l2_range_special.append([l2_range_special[-1][0] + " "])
        if len(key1_unique) > 8:
            l2_range_special.append([l2_range_special[-1][0] + " "])
        another_val = len(key1_unique) - 8
        if another_val > 0:
            for i in range(another_val):
                if i % 3 == 0:
                    l2_range_special.append([l2_range_special[-1][0] + " "])
        l2_xrange = pd.unique(key1_unique)
        l2_xrange2 = np.append(l2_xrange, l2_range_special)
        l1_title = y_axis_text + " over " + param_keys_title[0]
        l2_title = y_axis_text + " over " + param_keys_title[1]
        y_range_min = min(cv_score) - min(cv_score) * 0.1
        y_range_max = max(full_score) + max(full_score) * 0.05
        p6 = figure(title=l2_title, x_axis_label=param_keys_axis[1], y_axis_label=y_axis_text, plot_width=int(320 * scale), plot_height=int(257 * scale), x_range=l2_xrange2, tools="tap,pan,wheel_zoom,box_zoom,reset,save,lasso_select,box_select", y_range=(y_range_min, y_range_max))
        p6.quad(top=[1000], bottom=[-1000], left=[l2_xrange[-1]], right=[1000], color="white")

        if self.n_mc > 1:
            p6_render_patch2 = p6.patches("monte_line_key1_value", "monte_line0_cv", alpha=0, color="blue", source=source)
            p6_render_patch2.selection_glyph = Patches(fill_alpha=0.1, fill_color="blue", line_color="white")
            p6_render_patch2.nonselection_glyph.fill_alpha = 0
            p6_render_patch2.nonselection_glyph.line_color = "white"
            # kfold monte-carlo does not have ci for R2
            if model is not "kfold":
                p6_render_patch1 = p6.patches("monte_line_key1_value", "monte_line0_full", alpha=0, color="red", source=source)
                p6_render_patch1.selection_glyph = Patches(fill_alpha=0.1, fill_color="red", line_color="white")
                p6_render_patch1.nonselection_glyph.fill_alpha = 0
                p6_render_patch1.nonselection_glyph.line_color = "white"

        p6_render_1 = p6.multi_line("line_key1_value", "line0_full", line_color="red", line_width=2 * scale, source=source, legend=full_legend)
        p6_render_1.selection_glyph = MultiLine(line_color="red", line_alpha=0.8, line_width=2 * scale)
        p6_render_1.nonselection_glyph.line_color = "red"
        p6_render_1.nonselection_glyph.line_alpha = 0.05 / len(key0_unique)

        p6_render_2 = p6.circle("key0_value", "full_score", line_color="red", fill_color="white", size=8 * scale, source=source)
        p6_render_2.selection_glyph = Circle(line_color="red", fill_color="white")
        p6_render_2.nonselection_glyph.line_color = "red"
        p6_render_2.nonselection_glyph.fill_color = "white"
        p6_render_2.nonselection_glyph.line_alpha = 0.7 / len(key0_unique)

        p6_render_3 = p6.multi_line("line_key1_value", "line0_cv", line_color="blue", line_width=2 * scale, source=source, legend=cv_legend)
        p6_render_3.selection_glyph = MultiLine(line_color="blue", line_alpha=0.8, line_width=2 * scale)
        p6_render_3.nonselection_glyph.line_color = "blue"
        p6_render_3.nonselection_glyph.line_alpha = 0.05 / len(key0_unique)

        p6_render_4 = p6.circle("key0_value", "cv_score", line_color="blue", fill_color="white", size=8 * scale, source=source)
        p6_render_4.selection_glyph = Circle(line_color="blue", fill_color="white")
        p6_render_4.nonselection_glyph.line_color = "blue"
        p6_render_4.nonselection_glyph.fill_color = "white"
        p6_render_4.nonselection_glyph.line_alpha = 0.7 / len(key0_unique)

        # Text
        text_here = 8 * scale
        text_line_font = str(text_here) + "pt"
        p6_render_5 = p6.text(x="key0_value_text", y="line0_cv_text", text="line0_cv_score_text", source=source, text_font_size=text_line_font, text_color="blue", x_offset=8 * scale, y_offset=6 * scale, text_alpha=0)
        p6_render_5.selection_glyph = Text(text_color="blue", text_alpha=1, text_font_size=text_line_font)
        p6_render_5.nonselection_glyph.text_alpha = 0
        p6_render_6 = p6.text(x="key0_value_text", y="line0_full_text", text="line0_cv_score_text", source=source, text_font_size=text_line_font, text_color="red", x_offset=8 * scale, y_offset=6 * scale, text_alpha=0)
        p6_render_6.selection_glyph = Text(text_color="red", text_alpha=1, text_font_size=text_line_font)
        p6_render_6.nonselection_glyph.text_alpha = 0

        p6.add_tools(HoverTool(renderers=[p6_render_2], tooltips=[("AUC_full", "@full_text")]))

        p6.add_tools(HoverTool(renderers=[p6_render_4], tooltips=[("AUC_CV", "@cv_text")]))

        p6.add_tools(TapTool(renderers=[p6_render_2, p6_render_4]))

        # Figure: legend
        if legend == None or legend == False:
            p5.legend.visible = False
            p6.legend.visible = False
        else:
            p5.legend.location = legend
            p6.legend.location = legend

        # Center title
        if title_align == "center":
            p1.title.align = "center"
            p2.title.align = "center"
            p3.title.align = "center"
            p4.title.align = "center"
            p5.title.align = "center"
            p6.title.align = "center"

        fig = gridplot([[p1, p2, p3], [p4, p5, p6]], merge_tools=True, toolbar_location="left", toolbar_options=dict(logo=None))

        p5.xaxis.major_label_orientation = math.radians(heatmap_xaxis_rotate)
        p6.xaxis.major_label_orientation = math.radians(heatmap_xaxis_rotate)

        p1.title.text_font_size = str(12 * scale) + "pt"
        p2.title.text_font_size = str(12 * scale) + "pt"
        p3.title.text_font_size = str(12 * scale) + "pt"
        p4.title.text_font_size = str(12 * scale) + "pt"
        p5.title.text_font_size = str(12 * scale) + "pt"
        p6.title.text_font_size = str(12 * scale) + "pt"

        p1.xaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p2.xaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p3.xaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p4.xaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p5.xaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p6.xaxis.axis_label_text_font_size = str(10 * scale) + "pt"

        p1.yaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p2.yaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p3.yaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p4.yaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p5.yaxis.axis_label_text_font_size = str(10 * scale) + "pt"
        p6.yaxis.axis_label_text_font_size = str(10 * scale) + "pt"

        return fig



from keras.callbacks import Callback


class YpredCallback(Callback):
    """Used as a callback for Keras to get Ypred_train, and Ypred_test for each epoch.

    Example:
    yc = YpredCallback(X, X)
    model.fit(X, Y, callbacks=[yc]
    """

    def __init__(self, model, X_train, X_test=None):
        self.model = model  # Keras model
        self.Y_train = []
        self.Y_test = []
        self.X_train = X_train
        # If X_test is None, use X_train
        if X_test is None:
            self.X_test = X_train
        else:
            self.X_test = X_test

    def on_epoch_end(self, model, epoch, logs=None):
        Y_train_pred = self.model.predict(self.X_train).flatten()
        Y_test_pred = self.model.predict(self.X_test).flatten()
        self.Y_train.append(Y_train_pred)
        self.Y_test.append(Y_test_pred)



import numpy as np
import pandas as pd
from bokeh.models import Band, HoverTool
from tqdm import tqdm
import timeit
import time
import multiprocessing
from joblib import Parallel, delayed
from copy import deepcopy, copy
from bokeh.plotting import ColumnDataSource, figure
import scipy
from scipy import interp
from sklearn import metrics
from sklearn.metrics import confusion_matrix, roc_auc_score
from sklearn.utils import resample
from ..utils import binary_metrics, dict_median


def roc_plot(fpr, tpr, tpr_ci, median=False, width=450, height=350, xlabel="1-Specificity", ylabel="Sensitivity", legend=True, label_font_size="13pt", title="", errorbar=False):
    """Creates a rocplot using Bokeh.

    Parameters
    ----------
    fpr : array-like, shape = [n_samples]
        False positive rates. Calculate using roc_calculate.

    tpr : array-like, shape = [n_samples]
        True positive rates. Calculate using roc_calculate.

    tpr_ci : array-like, shape = [n_samples, 2]
        True positive rates 95% confidence intervals [lowci, uppci]. Calculate using roc_calculate.
    """

    # Get CI
    tpr_lowci = tpr_ci[0]
    tpr_uppci = tpr_ci[1]
    tpr_medci = tpr_ci[2]
    auc = metrics.auc(fpr, tpr)

    if median == True:
        tpr = tpr_medci

    # specificity and ci-interval for HoverTool
    spec = 1 - fpr
    ci = (tpr_uppci - tpr_lowci) / 2

    # Figure
    data = {"x": fpr, "y": tpr, "lowci": tpr_lowci, "uppci": tpr_uppci, "spec": spec, "ci": ci}
    source = ColumnDataSource(data=data)
    fig = figure(title=title, plot_width=width, plot_height=height, x_axis_label=xlabel, y_axis_label=ylabel, x_range=(-0.06, 1.06), y_range=(-0.06, 1.06))

    # Figure: add line
    fig.line([0, 1], [0, 1], color="black", line_dash="dashed", line_width=2.5, legend="Equal distribution line")
    figline = fig.line("x", "y", color="green", line_width=3.5, alpha=0.6, legend="ROC Curve (Train)", source=source)
    fig.add_tools(HoverTool(renderers=[figline], tooltips=[("Specificity", "@spec{1.111}"), ("Sensitivity", "@y{1.111} (+/- @ci{1.111})")]))

    # Figure: add 95CI band
    figband = Band(base="x", lower="lowci", upper="uppci", level="underlay", fill_alpha=0.1, line_width=1, line_color="black", fill_color="green", source=source)
    fig.add_layout(figband)

    # Figure: add errorbar  spec =  1 - fpr
    if errorbar is not False:
        idx = np.abs(fpr - (1 - errorbar)).argmin()  # this find the closest value in fpr to errorbar fpr
        fpr_eb = fpr[idx]
        tpr_eb = tpr[idx]
        tpr_lowci_eb = tpr_lowci[idx]
        tpr_uppci_eb = tpr_uppci[idx]

        # Edge case: If this is a perfect roc curve, and specificity >= 1, make sure error_bar is at (0,1) not (0,0)
        if errorbar >= 1:
            for i in range(len(fpr)):
                if fpr[i] == 0 and tpr[i] == 1:
                    fpr_eb = 0
                    tpr_eb = 1
                    tpr_lowci_eb = 1
                    tpr_uppci_eb = 1

        roc_whisker_line = fig.multi_line([[fpr_eb, fpr_eb]], [[tpr_lowci_eb, tpr_uppci_eb]], line_alpha=1, line_color="black")
        roc_whisker_bot = fig.multi_line([[fpr_eb - 0.03, fpr_eb + 0.03]], [[tpr_lowci_eb, tpr_lowci_eb]], line_color="black")
        roc_whisker_top = fig.multi_line([[fpr_eb - 0.03, fpr_eb + 0.03]], [[tpr_uppci_eb, tpr_uppci_eb]], line_alpha=1, line_color="black")
        fig.circle([fpr_eb], [tpr_eb], size=8, fill_alpha=1, line_alpha=1, line_color="black", fill_color="white")

    # Change font size
    fig.title.text_font_size = "11pt"
    fig.xaxis.axis_label_text_font_size = label_font_size
    fig.yaxis.axis_label_text_font_size = label_font_size
    fig.legend.label_text_font = "10pt"

    # Extra padding
    fig.min_border_left = 20
    fig.min_border_right = 20
    fig.min_border_top = 20
    fig.min_border_bottom = 20

    # Edit legend
    fig.legend.location = "bottom_right"
    fig.legend.label_text_font_size = "10pt"
    if legend is False:
        fig.legend.visible = False
    return fig


def roc_calculate(Ytrue, Yscore, bootnum=1000, metric=None, val=None, parametric=True):
    """Calculates required metrics for the roc plot function (fpr, tpr, and tpr_ci).

    Parameters
    ----------
    Ytrue : array-like, shape = [n_samples]
        Binary label for samples (0s and 1s)

    Yscore : array-like, shape = [n_samples]
        Predicted y score for samples

    Returns
    ----------------------------------
    fpr : array-like, shape = [n_samples]
        False positive rates.

    tpr : array-like, shape = [n_samples]
        True positive rates.

    tpr_ci : array-like, shape = [n_samples, 2]
        True positive rates 95% confidence intervals [lowci, uppci].
    """

    # Get fpr, tpr
    fpr, tpr, threshold = metrics.roc_curve(Ytrue, Yscore, pos_label=1, drop_intermediate=False)

    # fpr, tpr with drop_intermediates for fpr = 0 (useful for plot... since we plot specificity on x-axis, we don't need intermediates when fpr=0)
    tpr0 = tpr[fpr == 0][-1]
    tpr = np.concatenate([[tpr0], tpr[fpr > 0]])
    fpr = np.concatenate([[0], fpr[fpr > 0]])

    # if metric is provided, calculate stats
    if metric is not None:
        specificity, sensitivity, threshold = get_spec_sens_cuttoff(Ytrue, Yscore, metric, val)
        stats = get_stats(Ytrue, Yscore, specificity, parametric)
        stats["val_specificity"] = specificity
        stats["val_sensitivity"] = specificity
        stats["val_cutoffscore"] = threshold

    # bootstrap using vertical averaging
    tpr_boot = []
    boot_stats = []
    for i in range(bootnum):
        # Resample and get tpr, fpr
        Ytrue_res, Yscore_res = resample(Ytrue, Yscore)
        fpr_res, tpr_res, threshold_res = metrics.roc_curve(Ytrue_res, Yscore_res, pos_label=1, drop_intermediate=False)

        # Drop intermediates when fpr=0
        tpr0_res = tpr_res[fpr_res == 0][-1]
        tpr_res = np.concatenate([[tpr0_res], tpr_res[fpr_res > 0]])
        fpr_res = np.concatenate([[0], fpr_res[fpr_res > 0]])

        # Vertical averaging... use closest fpr_res to fpr, and append the corresponding tpr
        idx = [np.abs(i - fpr_res).argmin() for i in fpr]
        tpr_list = tpr_res[idx]
        tpr_boot.append(tpr_list)

        # if metric is provided, calculate stats
        if metric is not None:
            stats_res = get_stats(Ytrue_res, Yscore_res, specificity, parametric)
            boot_stats.append(stats_res)

    # Get CI for bootstat
    if metric is not None:
        bootci_stats = {}
        for i in boot_stats[0].keys():
            stats_i = [k[i] for k in boot_stats]
            stats_i = np.array(stats_i)
            stats_i = stats_i[~np.isnan(stats_i)]  # Remove nans
            try:
                lowci = np.percentile(stats_i, 2.5)
                uppci = np.percentile(stats_i, 97.5)
            except IndexError:
                lowci = np.nan
                uppci = np.nan
            bootci_stats[i] = [lowci, uppci]

    # Get CI for tpr
    tpr_lowci = np.percentile(tpr_boot, 2.5, axis=0)
    tpr_uppci = np.percentile(tpr_boot, 97.5, axis=0)
    tpr_medci = np.percentile(tpr_boot, 50, axis=0)

    # Add the starting 0
    tpr = np.insert(tpr, 0, 0)
    fpr = np.insert(fpr, 0, 0)
    tpr_lowci = np.insert(tpr_lowci, 0, 0)
    tpr_uppci = np.insert(tpr_uppci, 0, 0)
    tpr_medci = np.insert(tpr_medci, 0, 0)

    # Concatenate tpr_ci
    tpr_ci = np.array([tpr_lowci, tpr_uppci, tpr_medci])

    if metric is None:
        return fpr, tpr, tpr_ci
    else:
        return fpr, tpr, tpr_ci, stats, bootci_stats


def roc_plot_boot(fpr_ib, tpr_ib_ci, fpr_oob, tpr_oob_ci, width=450, height=350, xlabel="1-Specificity", ylabel="Sensitivity", legend=True, label_font_size="13pt", title="", errorbar=False):
    """Creates a rocplot using Bokeh.

    Parameters
    ----------
    fpr : array-like, shape = [n_samples]
        False positive rates. Calculate using roc_calculate.

    tpr : array-like, shape = [n_samples]
        True positive rates. Calculate using roc_calculate.

    tpr_ci : array-like, shape = [n_samples, 2]
        True positive rates 95% confidence intervals [lowci, uppci]. Calculate using roc_calculate.
    """

    # Get CI
    tpr_ib = tpr_ib_ci[0]
    tpr = tpr_ib
    tpr_ib_lowci = tpr_ib_ci[1]
    tpr_ib_uppci = tpr_ib_ci[2]
    fpr = np.insert(fpr_ib[0], 0, 0)
    auc_ib = metrics.auc(fpr, tpr_ib)

    # add oob median CI
    tpr_oob = tpr_oob_ci[0]
    tpr_oob_lowci = tpr_oob_ci[1]
    tpr_oob_uppci = tpr_oob_ci[2]
    auc_oob = metrics.auc(fpr, tpr_oob)

    # specificity and ci-interval for HoverTool
    spec = 1 - fpr
    ci = (tpr_oob_uppci - tpr_oob_lowci) / 2

    # Figure
    data = {"x": fpr, "y": tpr_ib, "lowci": tpr_ib_lowci, "uppci": tpr_ib_uppci, "spec": spec, "ci": ci, "x_oob": fpr, "y_oob": tpr_oob, "lowci_oob": tpr_oob_lowci, "uppci_oob": tpr_oob_uppci}
    source = ColumnDataSource(data=data)
    fig = figure(title=title, plot_width=width, plot_height=height, x_axis_label=xlabel, y_axis_label=ylabel, x_range=(-0.06, 1.06), y_range=(-0.06, 1.06))

    # Figure: add line
    fig.line([0, 1], [0, 1], color="black", line_dash="dashed", line_width=2.5, legend="Equal distribution line")
    figline = fig.line("x", "y", color="green", line_width=3.5, alpha=0.6, legend="ROC Curve (IB)", source=source)
    fig.add_tools(HoverTool(renderers=[figline], tooltips=[("Specificity", "@spec{1.111}"), ("Sensitivity", "@y{1.111} (+/- @ci{1.111})")]))

    # Figure: add 95CI band
    figband = Band(base="x", lower="lowci", upper="uppci", level="underlay", fill_alpha=0.1, line_width=1, line_color="black", fill_color="green", source=source)
    fig.add_layout(figband)

    # Figure: add oob line
    figline2 = fig.line("x_oob", "y_oob", color="red", line_width=3.5, alpha=0.6, legend="ROC Curve (OOB)", source=source)
    figband2 = Band(base="x_oob", lower="lowci_oob", upper="uppci_oob", level="underlay", fill_alpha=0.1, line_width=1, line_color="black", fill_color="red", source=source)
    fig.add_layout(figband2)

    # Figure: add errorbar  spec =  1 - fpr
    if errorbar is not False:
        idx = np.abs(fpr - (1 - errorbar)).argmin()  # this find the closest value in fpr to errorbar fpr
        fpr_eb = fpr[idx]
        tpr_eb = tpr[idx]
        tpr_lowci_eb = tpr_ib_lowci[idx]
        tpr_uppci_eb = tpr_ib_uppci[idx]

        # Edge case: If this is a perfect roc curve, and specificity >= 1, make sure error_bar is at (0,1) not (0,0)
        if errorbar >= 1:
            for i in range(len(fpr)):
                if fpr[i] == 0 and tpr[i] == 1:
                    fpr_eb = 0
                    tpr_eb = 1
                    tpr_lowci_eb = 1
                    tpr_uppci_eb = 1

        roc_whisker_line = fig.multi_line([[fpr_eb, fpr_eb]], [[tpr_lowci_eb, tpr_uppci_eb]], line_alpha=1, line_width=2, line_color="darkgreen")
        roc_whisker_bot = fig.multi_line([[fpr_eb - 0.03, fpr_eb + 0.03]], [[tpr_lowci_eb, tpr_lowci_eb]], line_width=2, line_color="darkgreen")
        roc_whisker_top = fig.multi_line([[fpr_eb - 0.03, fpr_eb + 0.03]], [[tpr_uppci_eb, tpr_uppci_eb]], line_width=2, line_alpha=1, line_color="darkgreen")
        fig.circle([fpr_eb], [tpr_eb], size=8, fill_alpha=1, line_alpha=1, line_color="black", fill_color="white")

    # Figure: add errorbar  spec =  1 - fpr
    if errorbar is not False:
        idx = np.abs(fpr - (1 - errorbar)).argmin()  # this find the closest value in fpr to errorbar fpr
        fpr_eb2 = fpr[idx]
        tpr_eb2 = tpr_oob[idx]
        tpr_lowci_eb2 = tpr_oob_lowci[idx]
        tpr_uppci_eb2 = tpr_oob_uppci[idx]

        # Edge case: If this is a perfect roc curve, and specificity >= 1, make sure error_bar is at (0,1) not (0,0)
        if errorbar >= 1:
            for i in range(len(fpr)):
                if fpr[i] == 0 and tpr_oob[i] == 1:
                    fpr_eb2 = 0
                    tpr_eb2 = 1
                    tpr_lowci_eb2 = 1
                    tpr_uppci_eb2 = 1

        roc_whisker_line2 = fig.multi_line([[fpr_eb2, fpr_eb2]], [[tpr_lowci_eb2, tpr_uppci_eb2]], line_width=2, line_alpha=1, line_color="darkred")
        roc_whisker_bot2 = fig.multi_line([[fpr_eb2 - 0.03, fpr_eb2 + 0.03]], [[tpr_lowci_eb2, tpr_lowci_eb2]], line_width=2, line_color="darkred")
        roc_whisker_top2 = fig.multi_line([[fpr_eb2 - 0.03, fpr_eb2 + 0.03]], [[tpr_uppci_eb2, tpr_uppci_eb2]], line_width=2, line_alpha=1, line_color="darkred")
        fig.circle([fpr_eb2], [tpr_eb2], size=8, fill_alpha=1, line_alpha=1, line_color="black", fill_color="white")

    # Change font size
    fig.title.text_font_size = "11pt"
    fig.xaxis.axis_label_text_font_size = label_font_size
    fig.yaxis.axis_label_text_font_size = label_font_size
    fig.legend.label_text_font = "10pt"

    # Extra padding
    fig.min_border_left = 20
    fig.min_border_right = 20
    fig.min_border_top = 20
    fig.min_border_bottom = 20

    # Edit legend
    fig.legend.location = "bottom_right"
    fig.legend.label_text_font_size = "10pt"
    if legend is False:
        fig.legend.visible = False
    return fig


def roc_calculate_boot(model, Xtrue, Ytrue, Yscore, bootnum=1000, metric=None, val=None, parametric=True, n_cores=-1):
    """Calculates required metrics for the roc plot function (fpr, tpr, and tpr_ci).

    Parameters
    ----------
    Ytrue : array-like, shape = [n_samples]
        Binary label for samples (0s and 1s)

    Yscore : array-like, shape = [n_samples]
        Predicted y score for samples

    Returns
    ----------------------------------
    fpr : array-like, shape = [n_samples]
        False positive rates.

    tpr : array-like, shape = [n_samples]
        True positive rates.

    tpr_ci : array-like, shape = [n_samples, 2]
        True positive rates 95% confidence intervals [lowci, uppci].
    """
    # if n_cores = -1, set n_cores to max_cores
    max_num_cores = multiprocessing.cpu_count()
    n_cores = n_cores
    if n_cores > max_num_cores:
        n_cores = -1
        print("Number of cores set too high. It will be set to the max number of cores in the system.", flush=True)
    if n_cores == -1:
        n_cores = max_num_cores
        print("Number of cores set to: {}".format(max_num_cores))

    time.sleep(0.5)  # Sleep for 0.5 secs to finish printing

    # Start Timer
    start = timeit.default_timer()

    # model copy
    model_boot = copy(model)

    # Get fpr, tpr
    fpr, tpr, threshold = metrics.roc_curve(Ytrue, Yscore, pos_label=1, drop_intermediate=False)

    # fpr, tpr with drop_intermediates for fpr = 0 (useful for plot... since we plot specificity on x-axis, we don't need intermediates when fpr=0)
    tpr0 = tpr[fpr == 0][-1]
    tpr = np.concatenate([[tpr0], tpr[fpr > 0]])
    fpr = np.concatenate([[0], fpr[fpr > 0]])

    # if metric is provided, calculate stats
    if metric is not None:
        specificity, sensitivity, threshold = get_spec_sens_cuttoff(Ytrue, Yscore, metric, val)
        stats = get_stats(Ytrue, Yscore, specificity, parametric)
        stats["val_specificity"] = specificity
        stats["val_sensitivity"] = specificity
        stats["val_cutoffscore"] = threshold

    # bootstrap using vertical averaging to linspace
    mean_fpr = np.linspace(0, 1, 1000)
    # understand location
    x_loc = pd.DataFrame(Xtrue)
    x0_loc = list(x_loc[Ytrue == 0].index)
    x1_loc = list(x_loc[Ytrue == 1].index)
    x_loc_ib_dict = {k: [] for k in list(x_loc.index)}
    x_loc_oob_dict = {k: [] for k in list(x_loc.index)}
    # stratified resample
    x0 = Xtrue[Ytrue == 0]
    x1 = Xtrue[Ytrue == 1]
    x0_idx = list(range(len(x0)))
    x1_idx = list(range(len(x1)))

    # input for parallel
    class para_class:
        def __init__(self, x0_idx, x1_idx, model_boot, x0, x1, metric, specificity, parametric, mean_fpr, n_cores, x0_loc, x1_loc):
            self.x0_idx = x0_idx
            self.x1_idx = x1_idx
            self.model_boot = model_boot
            self.x0 = x0
            self.x1 = x1
            self.metric = metric
            self.specificity = specificity
            self.parametric = parametric
            self.mean_fpr = mean_fpr
            self.n_cores = n_cores
            self.x0_loc = x0_loc
            self.x1_loc = x1_loc

        def _roc_calculate_boot_loop(self, i):
            val = _roc_calculate_boot_loop(self)
            return val

    self = para_class(x0_idx, x1_idx, model_boot, x0, x1, metric, specificity, parametric, mean_fpr, n_cores, x0_loc, x1_loc)

    para_output = Parallel(n_jobs=self.n_cores)(delayed(self._roc_calculate_boot_loop)(i) for i in tqdm(range(bootnum)))

    tpr_ib = []
    fpr_ib = []
    stat_ib_boot = []
    median_ib = []
    tpr_oob = []
    fpr_oob = []
    stat_oob_boot = []
    median_oob = []
    manw_pval = []
    temp_x_loc_ib_dict = []
    temp_x_loc_oob_dict = []

    for i in para_output:
        tpr_ib.append(i[0])
        fpr_ib.append(i[1])
        stat_ib_boot.append(i[2])
        median_ib.append(i[3])
        tpr_oob.append(i[4])
        fpr_oob.append(i[5])
        stat_oob_boot.append(i[6])
        median_oob.append(i[7])
        manw_pval.append(i[8])
        temp_x_loc_ib_dict.append(i[9])
        temp_x_loc_oob_dict.append(i[10])

    for i in temp_x_loc_ib_dict:
        for j in i:
            x_loc_ib_dict[j[0]].append(j[1])

    for i in temp_x_loc_oob_dict:
        for j in i:
            x_loc_oob_dict[j[0]].append(j[1])

    # Get CI for bootstat ib
    if metric is not None:
        stat_ib = {}
        for i in stat_ib_boot[0].keys():
            stats_i = [k[i] for k in stat_ib_boot]
            stats_i = np.array(stats_i)
            stats_i = stats_i[~np.isnan(stats_i)]  # Remove nans
            try:
                lowci = np.percentile(stats_i, 2.5)
                medci = np.percentile(stats_i, 50)
                uppci = np.percentile(stats_i, 97.5)
            except IndexError:
                lowci = np.nan
                medci = np.nan
                uppci = np.nan
            stat_ib[i] = [medci, lowci, uppci]

    # Get CI for bootstat oob
    if metric is not None:
        stat_oob = {}
        for i in stat_oob_boot[0].keys():
            stats_i = [k[i] for k in stat_oob_boot]
            stats_i = np.array(stats_i)
            stats_i = stats_i[~np.isnan(stats_i)]  # Remove nans
            try:
                lowci = np.percentile(stats_i, 2.5)
                medci = np.percentile(stats_i, 50)
                uppci = np.percentile(stats_i, 97.5)
            except IndexError:
                lowci = np.nan
                medci = np.nan
                uppci = np.nan
            stat_oob[i] = [medci, lowci, uppci]

    # Get CI for tpr
    tpr_oob_lowci = np.percentile(tpr_oob, 2.5, axis=0)
    tpr_oob_medci = np.percentile(tpr_oob, 50, axis=0)
    tpr_oob_uppci = np.percentile(tpr_oob, 97.5, axis=0)

    # Add the starting 0
    tpr_oob_lowci = np.insert(tpr_oob_lowci, 0, 0)
    tpr_oob_medci = np.insert(tpr_oob_medci, 0, 0)
    tpr_oob_uppci = np.insert(tpr_oob_uppci, 0, 0)

    # Concatenate tpr_ci
    tpr_oob_ci = np.array([tpr_oob_medci, tpr_oob_lowci, tpr_oob_uppci])

    # Get CI for tpr
    tpr_ib_lowci = np.percentile(tpr_ib, 2.5, axis=0)
    tpr_ib_medci = np.percentile(tpr_ib, 50, axis=0)
    tpr_ib_uppci = np.percentile(tpr_ib, 97.5, axis=0)

    # Add the starting 0
    tpr_ib_lowci = np.insert(tpr_ib_lowci, 0, 0)
    tpr_ib_medci = np.insert(tpr_ib_medci, 0, 0)
    tpr_ib_uppci = np.insert(tpr_ib_uppci, 0, 0)

    # Get median score per boot
    median_y_ib = x_loc_ib_dict
    median_y_oob = x_loc_oob_dict

    # Concatenate tpr_ci
    tpr_ib_ci = np.array([tpr_ib_medci, tpr_ib_lowci, tpr_ib_uppci])

    # Stop timer
    stop = timeit.default_timer()
    self.parallel_time = (stop - start) / 60
    print("Time taken: {:0.2f} minutes with {} cores".format(self.parallel_time, self.n_cores))

    if metric is None:
        return fpr, tpr, tpr_ci
    else:
        return fpr_ib, tpr_ib_ci, stat_ib, median_ib, fpr_oob, tpr_oob_ci, stat_oob, median_oob, stats, median_y_ib, median_y_oob, manw_pval


def _roc_calculate_boot_loop(self):
    "loop using joblib"

    x0_idx = self.x0_idx
    x1_idx = self.x1_idx
    model_boot = self.model_boot
    x0 = self.x0
    x1 = self.x1
    metric = self.metric
    specificity = self.specificity
    parametric = self.parametric
    mean_fpr = self.mean_fpr
    x0_loc = self.x0_loc
    x1_loc = self.x1_loc

    # resample
    x0_idx_ib = resample(x0_idx)
    x1_idx_ib = resample(x1_idx)
    x0_idx_oob = list(set(x0_idx) - set(x0_idx_ib))
    x1_idx_oob = list(set(x1_idx) - set(x1_idx_ib))
    # get x
    x0_ib = x0[x0_idx_ib]
    x1_ib = x1[x1_idx_ib]
    x0_oob = x0[x0_idx_oob]
    x1_oob = x1[x1_idx_oob]
    x_ib = np.concatenate((x0_ib, x1_ib))
    x_oob = np.concatenate((x0_oob, x1_oob))
    # get y
    y0_ib = np.zeros(len(x0_idx_ib))
    y1_ib = np.ones(len(x1_idx_ib))
    y0_oob = np.zeros(len(x0_idx_oob))
    y1_oob = np.ones(len(x1_idx_oob))
    y_ib = np.concatenate((y0_ib, y1_ib))
    y_oob = np.concatenate((y0_oob, y1_oob))
    # train and test model
    ypred_ib = model_boot.train(x_ib, y_ib)
    ypred_oob = model_boot.test(x_oob)
    # get median ypred per group
    ypred_ib_0 = ypred_ib[: len(x0_idx_ib)]
    ypred_ib_1 = ypred_ib[len(x0_idx_ib):]
    k_median_ib = [np.median(ypred_ib_0), np.median(ypred_ib_1)]  # 1

    # get ib fpr, tpr, stats
    fpri, tpri, _ = metrics.roc_curve(y_ib, ypred_ib, pos_label=1, drop_intermediate=False)
    k_fpr_ib = mean_fpr
    k_tpr_ib = interp(mean_fpr, fpri, tpri)

    # tpr_ib[-1][0] = 0.0
    # if metric is provided, calculate stats
    if metric is not None:
        stats_resi = get_stats(y_ib, ypred_ib, specificity, parametric)
        k_stat_ib_boot = stats_resi

    # get median ypred per group
    ypred_oob_0 = ypred_oob[: len(x0_idx_oob)]
    ypred_oob_1 = ypred_oob[len(x0_idx_oob):]
    k_median_oob = [np.median(ypred_oob_0), np.median(ypred_oob_1)]

    # get oob
    fpro, tpro, _ = metrics.roc_curve(y_oob, ypred_oob, pos_label=1, drop_intermediate=False)
    k_fpr_oob = mean_fpr
    k_tpr_oob = interp(mean_fpr, fpro, tpro)
    #k_tpr_oob[-1][0] = 0.0

    # if metric is provided, calculate stats
    if metric is not None:
        stats_reso = get_stats(y_oob, ypred_oob, specificity, parametric)
        k_stat_oob_boot = stats_reso
    # manu
    try:
        manw_pval_ib = scipy.stats.mannwhitneyu(ypred_ib_0, ypred_ib_1, alternative="two-sided")[1]
    except ValueError:
        manw_pval_ib = 0
    try:
        manw_pval_oob = scipy.stats.mannwhitneyu(ypred_oob_0, ypred_oob_1, alternative="two-sided")[1]
    except ValueError:
        manw_pval_oob = 0
    k_manw_pval = [manw_pval_ib, manw_pval_oob]

    # get average ypred
    k_x_loc_ib_dict = []
    for i in range(len(ypred_ib_0)):
        idx_res = x0_idx_ib[i]
        idx_true = x0_loc[idx_res]
        k_x_loc_ib_dict.append([idx_true, ypred_ib_0[i]])
    for i in range(len(ypred_ib_1)):
        idx_res = x1_idx_ib[i]
        idx_true = x1_loc[idx_res]
        k_x_loc_ib_dict.append([idx_true, ypred_ib_1[i]])

    # get average ypred
    k_x_loc_oob_dict = []
    for i in range(len(ypred_oob_0)):
        idx_res = x0_idx_oob[i]
        idx_true = x0_loc[idx_res]
        k_x_loc_oob_dict.append([idx_true, ypred_oob_0[i]])
    for i in range(len(ypred_oob_1)):
        idx_res = x1_idx_oob[i]
        idx_true = x1_loc[idx_res]
        k_x_loc_oob_dict.append([idx_true, ypred_oob_1[i]])

    return [k_tpr_ib, k_fpr_ib, k_stat_ib_boot, k_median_ib, k_tpr_oob, k_fpr_oob, k_stat_oob_boot, k_median_oob, k_manw_pval, k_x_loc_ib_dict, k_x_loc_oob_dict]


def get_sens_spec(Ytrue, Yscore, cuttoff_val):
    """Get sensitivity and specificity from cutoff value."""
    Yscore_round = np.where(np.array(Yscore) > cuttoff_val, 1, 0)
    tn, fp, fn, tp = metrics.confusion_matrix(Ytrue, Yscore_round).ravel()
    sensitivity = tp / (tp + fn)
    specificity = tn / (tn + fp)
    return sensitivity, specificity


def get_sens_cuttoff(Ytrue, Yscore, specificity_val):
    """Get sensitivity and cuttoff value from specificity."""
    fpr0 = 1 - specificity_val
    fpr, sensitivity, thresholds = metrics.roc_curve(Ytrue, Yscore, pos_label=1, drop_intermediate=False)
    idx = np.abs(fpr - fpr0).argmin()  # this find the closest value in fpr to fpr0
    # Check that this is not a perfect roc curve
    # If it is perfect, allow sensitivity = 1, rather than 0
    if specificity_val == 1 and sensitivity[idx] == 0:
        for i in range(len(fpr)):
            if fpr[i] == 1 and sensitivity[i] == 1:
                return 1, 0.5
    return sensitivity[idx], thresholds[idx]


def get_spec_sens_cuttoff(Ytrue, Yscore, metric, val):
    """Return specificity, sensitivity, cutoff value provided the metric and value used."""
    if metric == "specificity":
        specificity = val
        sensitivity, threshold = get_sens_cuttoff(Ytrue, Yscore, val)
    elif metric == "cutoffscore":
        threshold = val
        sensitivity, specificity = get_sens_spec(Ytrue, Yscore, val)
    return specificity, sensitivity, threshold


def get_stats(Ytrue, Yscore, specificity, parametric):
    """Calculates binary metrics given the specificity."""
    sensitivity, cutoffscore = get_sens_cuttoff(Ytrue, Yscore, specificity)
    stats = binary_metrics(Ytrue, Yscore, cut_off=cutoffscore, parametric=parametric)
    return stats



from .binary_metrics import binary_metrics
from .ci95_ellipse import ci95_ellipse
from .dict_mean import dict_mean
from .dict_median import dict_median
from .dict_std import dict_std
from .dict_perc import dict_perc
from .knnimpute import knnimpute
from .load_comparisonXL import load_comparisonXL
from .load_dataXL import load_dataXL
from .scale import scale
from .nested_getattr import nested_getattr
from .table_check import table_check
from .univariate_2class import univariate_2class
from .wmean import wmean
from .YpredCallback import YpredCallback
from .color_scale import color_scale

__all__ = ["binary_metrics", "ci95_ellipse", "dict_mean", "dict_median", "dict_std", "dict_perc", "knnimpute", "load_comparisonXL", "load_dataXL", "scale", "nested_getattr", "table_check", "univariate_2class", "wmean", "YpredCallback", "color_scale"]



import numpy as np
import pandas as pd
from sklearn.svm import SVC
from sklearn.metrics import roc_auc_score
from .BaseModel import BaseModel


class SVM(BaseModel):
    """Support Vector Machine"""

    parametric = True
    bootlist = None  # list of metrics to bootstrap

    def __init__(self, C=1.0, kernel="rbf", degree=3, gamma="auto", tol=0.001, max_iter=-1):
        self.model = SVC(C=C, kernel=kernel, degree=degree, gamma=gamma, probability=True, tol=tol, max_iter=max_iter)
        self.k = None

    def set_params(self, params):
        self.__init__(**params)

    def train(self, X, Y):
        """ Fit the SVM model, save additional stats (as attributes) and return Y predicted values.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.

        Y : array-like, shape = [n_samples, 1]
            Response variables, where n_samples is the number of samples.

        Returns
        -------
        y_pred_train : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # Ensure array and error check
        X, Y = self.input_check(X, Y)

        # Fit the model
        self.model.fit(X, Y)

        # Predict_proba was designed for multi-groups...
        # This makes it sure that y_pred is correct
        y_pred = self.model.predict_proba(X)
        pred_0 = roc_auc_score(Y, y_pred[:, 0])
        pred_1 = roc_auc_score(Y, y_pred[:, 1])
        if pred_0 > pred_1:
            self.pred_index = 0
        else:
            self.pred_index = 1

        # Calculate and return Y prediction value
        y_pred_train = np.array(self.model.predict_proba(X)[:, self.pred_index])

        # Storing X, Y, and Y_pred
        self.X = X
        self.Y = Y
        self.Y_pred = y_pred_train
        return y_pred_train

    def test(self, X):
        """Calculate and return Y predicted value.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Test variables, where n_samples is the number of samples and n_features is the number of predictors.

        Returns
        -------
        y_pred_test : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # Convert to X to numpy array if a DataFrame
        if isinstance(X, pd.DataFrame or pd.Series):
            X = np.array(X)

        # Calculate and return Y predicted value
        y_pred_test = np.array(self.model.predict_proba(X)[:, self.pred_index])
        return y_pred_test



import numpy as np


def dict_mean(dict_list):
    mean_dict = {}
    for key in dict_list[0].keys():
        value = []
        for i in dict_list:
            value.append(i[key])
        mean_dict[key] = np.mean(value)
    return mean_dict



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="cimcb",
    version="1.0.20",
    description="This is a pre-release.",
    long_description=long_description,
    license="http://www.apache.org/licenses/LICENSE-2.0.html",
    url="https://github.com/KevinMMendez/cimcb",
    packages=["cimcb", "cimcb.bootstrap", "cimcb.cross_val", "cimcb.model", "cimcb.plot", "cimcb.utils"],
    python_requires=">=3.5",
    install_requires=["bokeh>=1.0.0",
                      "keras",
                      "numpy>=1.12",
                      "pandas",
                      "scipy",
                      "scikit-learn",
                      "statsmodels",
                      "tensorflow",
                      "tqdm",
                      "xlrd",
                      "joblib"],
    author="Kevin Mendez, David Broadhurst",
    author_email="k.mendez@ecu.edu.au, d.broadhurst@ecu.edu.au",
)



import numpy as np
from keras.callbacks import Callback
from keras.optimizers import SGD
from keras.models import Sequential
from keras.layers import Dense
import tensorflow as tf
from keras import backend as K
from .BaseModel import BaseModel
from ..utils import YpredCallback


class NN_LinearTan(BaseModel):
    """2 Layer linear-logistic neural network using Keras"""

    parametric = True
    bootlist = None

    def __init__(self, n_neurons=2, epochs=200, learning_rate=0.01, momentum=0.0, decay=0.0, nesterov=False, loss="binary_crossentropy", batch_size=None, verbose=0):
        self.n_neurons = n_neurons
        self.verbose = verbose
        self.n_epochs = epochs
        self.k = n_neurons
        self.batch_size = batch_size
        self.loss = loss
        self.optimizer = SGD(lr=learning_rate, momentum=momentum, decay=decay, nesterov=nesterov)

    def set_params(self, params):
        self.__init__(**params)

    def train(self, X, Y, epoch_ypred=False, epoch_xtest=None):
        """ Fit the neural network model, save additional stats (as attributes) and return Y predicted values.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.

        Y : array-like, shape = [n_samples, 1]
            Response variables, where n_samples is the number of samples.

        Returns
        -------
        y_pred_train : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        # # If using Keras, set tf to 1 core
        # config = K.tf.ConfigProto(intra_op_parallelism_threads=8, inter_op_parallelism_threads=8, allow_soft_placement=True)
        # session = tf.Session(config=config)
        # K.set_session(session)

        # If batch-size is None:
        if self.batch_size is None:
            self.batch_size = len(X)

        self.model = Sequential()
        self.model.add(Dense(self.n_neurons, activation="linear", input_dim=len(X.T)))
        self.model.add(Dense(1, activation="tanh"))
        self.model.compile(optimizer=self.optimizer, loss=self.loss, metrics=["accuracy"])

        # If epoch_ypred is True, calculate ypred for each epoch
        if epoch_ypred is True:
            self.epoch = YpredCallback(self.model, X, epoch_xtest)
        else:
            self.epoch = Callback()

        # Fit
        self.model.fit(X, Y, epochs=self.n_epochs, batch_size=self.batch_size, verbose=self.verbose, callbacks=[self.epoch])

        layer1_weight = self.model.layers[0].get_weights()[0]
        layer1_bias = self.model.layers[0].get_weights()[1]
        layer2_weight = self.model.layers[1].get_weights()[0]
        layer2_bias = self.model.layers[1].get_weights()[1]

        # Not sure about the naming scheme (trying to match PLS)
        self.model.x_loadings_ = layer1_weight
        self.model.x_scores_ = np.matmul(X, self.model.x_loadings_) + layer1_bias
        self.model.y_loadings_ = layer2_weight
        self.model.pctvar_ = np.ones((1, len(self.model.y_loadings_[0])))
        self.xcols_num = len(X.T)
        self.model.pctvar_ = sum(abs(self.model.x_scores_) ** 2) / sum(sum(abs(X) ** 2)) * 100
        y_pred_train = self.model.predict(X).flatten()

        # Storing X, Y, and Y_pred
        self.Y_pred = y_pred_train
        self.X = X
        self.Y = Y
        return y_pred_train

    def test(self, X, Y=None):
        """Calculate and return Y predicted value.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            Test variables, where n_samples is the number of samples and n_features is the number of predictors.

        Returns
        -------
        y_pred_test : array-like, shape = [n_samples, 1]
            Predicted y score for samples.
        """

        layer1_weight = self.model.layers[0].get_weights()[0]
        layer1_bias = self.model.layers[0].get_weights()[1]
        layer2_weight = self.model.layers[1].get_weights()[0]
        layer2_bias = self.model.layers[1].get_weights()[1]

        self.model.y_loadings_ = layer2_weight
        self.model.y_loadings_ = self.model.y_loadings_.reshape(1, len(self.model.y_loadings_))
        self.model.x_loadings_ = layer1_weight
        self.model.x_scores_ = np.matmul(X, self.model.x_loadings_) + layer1_bias
        y_pred_test = self.model.predict(X).flatten()
        self.model.pctvar_ = sum(abs(self.model.x_scores_) ** 2) / sum(sum(abs(X) ** 2)) * 100
        return y_pred_test



import numpy as np
from joblib import Parallel, delayed
import timeit
import time
from tqdm import tqdm
from sklearn.model_selection import train_test_split
from .BaseCrossVal import BaseCrossVal
from ..utils import binary_metrics, dict_perc, dict_median


class holdout(BaseCrossVal):
    """ Exhaustitive search over param_dict calculating binary metrics.

    Parameters
    ----------
    model : object
        This object is assumed to store bootlist attributes in .model (e.g. modelPLS.model.x_scores_).

    X : array-like, shape = [n_samples, n_features]
        Predictor variables, where n_samples is the number of samples and n_features is the number of predictors.

    Y : array-like, shape = [n_samples, 1]
        Response variables, where n_samples is the number of samples.

    param_dict : dict
        List of attributes to calculate and return bootstrap confidence intervals.

    split: : a positive number between 0 and 1, (default 0.8)
        The split for the train and test split.

    Methods
    -------
    Run: Runs all necessary methods prior to plot.

    Plot: Creates a R2/Q2 plot.
    """

    def __init__(self, model, X, Y, param_dict, folds=None, n_mc=1, n_boot=0, n_cores=-1, ci=95, test_size=0.2, stratify=True):

        super().__init__(model=model, X=X, Y=Y, param_dict=param_dict, folds=folds, n_mc=n_mc, n_boot=n_boot, n_cores=n_cores, ci=ci)

        # Save holdout specific inputs
        self.test_size = test_size
        if stratify is True:
            self.stratify = Y
        else:
            self.stratify = None

    def calc_ypred(self):
        """Calculates ypred full and ypred cv."""

        time.sleep(0.5)  # Sleep for 0.5 secs to finish printing

        # Start Timer
        start = timeit.default_timer()

        # Actual loop including Monte-Carlo reps
        self.loop_mc = self.param_list * self.n_mc
        ypred = Parallel(n_jobs=self.n_cores)(delayed(self._calc_ypred_loop)(i) for i in tqdm(range(len(self.loop_mc))))

        # Split ypred into full / cv and put in final format
        # Format :::> self.ypred_full -> parameter_type -> monte-carlo -> y_true / y_pred
        self.ypred_full = [[] for i in range(len(self.param_list))]
        self.ypred_cv = [[] for i in range(len(self.param_list))]
        self.loop_mc_numbers = list(range(len(self.param_list))) * self.n_mc
        for i in range(len(self.loop_mc)):
            j = self.loop_mc_numbers[i]  # Location to append to
            self.ypred_full[j].append(ypred[i][0])
            self.ypred_cv[j].append(ypred[i][1])

        # Stop timer
        stop = timeit.default_timer()
        self.parallel_time = (stop - start) / 60
        print("Time taken: {:0.2f} minutes with {} cores".format(self.parallel_time, self.n_cores))

    def calc_ypred_epoch(self):
        """Calculates ypred full and ypred cv for each epoch (edge case)."""

        time.sleep(0.5)  # Sleep for 0.5 secs to finish printing

        # Start Timer
        start = timeit.default_timer()

        # Set param to the max -> Actual loop including Monte-Carlo reps
        epoch_param = [self.param_list[-1]]
        self.loop_mc = epoch_param * self.n_mc
        ypred = Parallel(n_jobs=self.n_cores)(delayed(self._calc_ypred_loop_epoch)(i) for i in tqdm(range(len(self.loop_mc))))

        # Get epoch list
        self.epoch_list = []
        for m in self.param_list2:
            for t, v in m.items():
                self.epoch_list.append(v - 1)

        # Split ypred into full / cv and put in final format
        # Format :::> self.ypred_full -> parameter_type -> monte-carlo -> y_true / y_pred
        # Note, we need to pull out the specific epochs from the model
        self.ypred_full = [[] for i in range(len(self.epoch_list))]
        self.ypred_cv = [[] for i in range(len(self.epoch_list))]
        for i in range(len(self.loop_mc)):
            for j in range(len(self.epoch_list)):
                actual_epoch = self.epoch_list[j]
                self.ypred_full[j].append([ypred[i][0][0][0], ypred[i][0][0][1][actual_epoch]])
                self.ypred_cv[j].append([ypred[i][1][0][0], ypred[i][1][0][1][actual_epoch]])

        # Stop timer
        stop = timeit.default_timer()
        self.parallel_time = (stop - start) / 60
        print("Time taken: {:0.2f} minutes with {} cores".format(self.parallel_time, self.n_cores))

    def _calc_ypred_loop(self, i):
        """Core component of calc_ypred."""
        # Set x and y
        X_train, X_test, Y_train, Y_test = train_test_split(self.X, self.Y, test_size=self.test_size, stratify=self.stratify)
        # Set hyper - parameters
        params_i = self.loop_mc[i]
        model_i = self.model()
        model_i.set_params(params_i)
        # Split into train and test
        ypred_full_i = model_i.train(X_train, Y_train)
        ypred_cv_i = model_i.test(X_test)
        # Get ypred full cv
        ypred_full = [Y_train, ypred_full_i]
        ypred_cv = [Y_test, ypred_cv_i]
        return [ypred_full, ypred_cv]

    def _calc_ypred_loop_epoch(self, i):
        """Core component of calc_ypred_epoch."""
        # Set inputs
        Y_full = []
        Y_cv = []
        X_train, X_test, Y_train, Y_test = train_test_split(self.X, self.Y, test_size=self.test_size, stratify=self.stratify)
        # Set hyper - parameters
        params_i = self.loop_mc[i]
        model_i = self.model()
        model_i.set_params(params_i)
        # Train model with epoch_ypred=True
        model_i.train(X_train, Y_train, epoch_ypred=True, epoch_xtest=X_test)
        Y_full_split = model_i.epoch.Y_train
        Y_full.append([Y_train, Y_full_split])
        Y_cv_split = model_i.epoch.Y_test
        Y_cv.append([Y_test, Y_cv_split])
        return [Y_full, Y_cv]

    def calc_stats(self):
        """Calculates binary statistics from ypred full and ypred cv."""
        # Calculate for each parameter and append
        stats_list = []
        std_list = []
        self.full_loop = []
        self.cv_loop = []
        for i in range(len(self.param_list)):
            full_loop = []
            cv_loop = []

            # Get all monte-carlo
            for k in range(len(self.ypred_full[i])):
                full_mc = binary_metrics(self.ypred_full[i][k][0], self.ypred_full[i][k][1], parametric=self.model.parametric)
                cv_mc = binary_metrics(self.ypred_cv[i][k][0], self.ypred_cv[i][k][1], parametric=self.model.parametric)
                full_loop.append(full_mc)
                cv_loop.append(cv_mc)

            # Average binary metrics
            stats_full_i = dict_median(full_loop)
            stats_cv_i = dict_median(cv_loop)

            # Rename columns
            stats_full_i = {k + "full": v for k, v in stats_full_i.items()}
            stats_cv_i = {k + "cv": v for k, v in stats_cv_i.items()}
            stats_cv_i["R²"] = stats_full_i.pop("R²full")
            stats_cv_i["Q²"] = stats_cv_i.pop("R²cv")

            # Combine and append
            stats_combined = {**stats_full_i, **stats_cv_i}
            stats_list.append(stats_combined)

            # Save loop -> full_loop is a placeholder
            self.full_loop.append(full_loop)
            self.cv_loop.append(cv_loop)

            # Keep std if n_mc > 1
            if self.n_mc > 1:
                std_full_i = dict_perc(full_loop, ci=self.ci)
                std_cv_i = dict_perc(cv_loop, ci=self.ci)
                std_full_i = {k + "full": v for k, v in std_full_i.items()}
                std_cv_i = {k + "cv": v for k, v in std_cv_i.items()}
                std_cv_i["R²"] = std_full_i.pop("R²full")
                std_cv_i["Q²"] = std_cv_i.pop("R²cv")
                std_combined = {**std_full_i, **std_cv_i}
                std_list.append(std_combined)

        self.table = self._format_table(stats_list)  # Transpose, Add headers
        if self.n_mc > 1:
            self.table_std = self._format_table(std_list)  # Transpose, Add headers
        return self.table

    def plot(self, metric="r2q2", scale=1, color_scaling="tanh", rotate_xlabel=True, legend="bottom_right", color_beta=[10, 10, 10], ci=95):
        super().plot(metric=metric, scale=scale, color_scaling=color_scaling, rotate_xlabel=rotate_xlabel, legend=legend, model="holdout", color_beta=color_beta, ci=ci)



import pandas as pd


def load_comparisonXL(method, evaluate="train", dropna=True):
    """Load comparison table."""
    if evaluate == "test":
        e = "['Test']"
    elif evaluate == "in bag":
        e = "['In Bag']"
    elif evaluate == "out of bag":
        e = "['Out of Bag']"
    else:
        e = "['Train']"

    # Import methods
    table = []
    for i in method:
        table.append(pd.read_excel(i + ".xlsx"))

    # Concatenate table
    df = pd.DataFrame()
    for i in range(len(table)):
        df = pd.concat([df, table[i].loc[table[i]['evaluate'] == e].T.squeeze()], axis=1, sort=False)
    df = df.T.drop(columns="evaluate")

    # Remove [ ] from string
    for i in range(len(df)):
        for j in range(len(df.T)):
            if type(df.iloc[i, j]) is str:
                df.iloc[i, j] = df.iloc[i, j][2: -2]

    # Reset index and add methods column
    method_name = []
    for i in range(len(method)):
        name_i = method[i].rsplit('/', 1)[1]
        method_name.append(name_i)
    df = df.reset_index()
    df = pd.concat([pd.Series(method_name, name="method"), df], axis=1, sort=False)
    df = df.drop("index", 1)
    #df = df.set_index("method")

    # drop columns with just nans
    if dropna is True:
        df = df.dropna(axis=1, how='all')

    return df
